"""contains the copulAX implementation of a univariate fitter object."""
import jax.numpy as jnp
from jax import jit
from typing import Iterable
from functools import partial
from collections import deque

from copulax.univariate.distributions import *
from copulax.univariate.distributions import _dist_tree
from copulax._src.typing import Scalar
from copulax._src._distributions import Univariate


def _get_dist_objects(dists: Iterable | str) -> tuple:
    if isinstance(dists, str):
        dists: str = dists.lower().strip()
        if dists not in ("all", "common", "continuous", "discrete", "common continuous", "common discrete"):
            raise ValueError(f"Invalid value for 'dists' argument: {dists}." \
                             "If a string, dists must be one of 'all', " \
                                "'common', 'continuous', 'discrete', " \
                                "'common continuous' or 'common disrete'.")
        
        elif dists == "all":
            dists_objs: tuple = (*_dist_tree["continuous"].values(), 
                                 *_dist_tree["discrete"].values())
        elif dists in  ("common continuous", "common discrete"):
            dists_objs: tuple = tuple(_dist_tree["common"][dists.split()[-1]].values())
        elif dists == "common":
            dists_objs: tuple = tuple((
                *_dist_tree["common"]["continuous"].values(), 
                *_dist_tree["common"]["discrete"].values()))
        else:
            dists_objs: tuple = tuple(_dist_tree[dists].values())

    elif isinstance(dists, Iterable):
        dists_objs: tuple = tuple(dists)
        for dist in dists:
            if not isinstance(dist, Univariate):
                raise ValueError(f"Invalid distribution object provided " \
                                 f"within 'dists' iterable: {dist}. " \
                                 f"Distribution objects must be univariate " \
                                 "copulax distributions.")
    else:
        raise ValueError(f"Invalid value for 'dists' argument: {dists}. " \
                         "Dists must be a string or an iterable of " \
                         "copulAX distribution objects.")
    
    return dists_objs


@partial(jit, static_argnames=('metric',))
def _fit_and_stats(dist, x, metric, **kwargs):
    dist_params = dist.fit(x, **kwargs)
    dist_metric = getattr(dist, metric)(x=x, params=dist_params)
    return {'params': dist_params, 'metric': dist_metric, 'dist': dist, }


# @partial(jit, static_argnames=('metric', 'distributions'))
def univariate_fitter(x: jnp.ndarray, metric: str = "bic", 
                      distributions: Iterable | str = "common continuous", 
                      **kwargs) -> dict:
    r"""Find and fit the 'best' univariate distribution to the input data 
    according to a specified metric.

    Note:
        If you intend to jit wrap this function, ensure that 'metric' 
        and 'distributions' are static arguments.

    Args:
        x (ArrayLike): The input data to fit a distribution to.
        metric (str): The metric to use when selecting the 'best' distribution. 
        Must be one of 'aic', 'bic' or 'loglikelihood'. Default is 'bic'.
        distributions (Iterable | str): The distribution(s) to fit to the data. If a 
        string, must be one of 'all', 'common', 'continuous', 'discrete', 
        'common continuous' or 'common discrete' corresponding to the 
        classifications within copulax.univariate.distributions.distributions. 
        If an iterable, must contain copulAX distribution objects. Default is 
        'common continuous'.
        kwargs: Additional keyword arguments to pass to each distribution 
        object's fit method.

    Returns:
        res (tuple): The index of the best distribution fit and a tuple of 
        fitted distributions and the metric values.

    Examples:
    >>> import jax.numpy as jnp
    >>> import numpy as np
    >>> from copulax.univariate import univariate_fitter
    >>> x = np.random.normal(0, 1, 100)
    >>> univariate_fitter(x)
    """
    # getting distribution objects
    dists_objs = _get_dist_objects(distributions)

    # checking metric
    if metric not in ("aic", "bic", "loglikelihood"):
        raise ValueError(f"Invalid value for 'metric' argument: {metric}. " \
                         "Must be one of 'aic', 'bic' or 'loglikelihood'.")

    # fitting distributions
    output: deque = deque(maxlen=len(dists_objs))
    best_index, best_metric = 0, jnp.inf
    for i, dist in enumerate(dists_objs):
        dist_output: dict = _fit_and_stats(dist=dist, x=x, metric=metric, **kwargs)
        dist_metric: Scalar = dist_output["metric"]
        cond: bool = dist_metric < best_metric
        best_index: int = jnp.where(cond, i, best_index)
        best_metric: Scalar = jnp.where(cond, dist_metric, best_metric)
        
        output.append(dist_output)

    return best_index, tuple(output)
 


"""File containing the copulAX implementation of various covariance 
matrices."""
from jax.tree_util import register_pytree_node
import jax.numpy as jnp
from jax import lax, random, jit
from jax._src.typing import ArrayLike, Array
import jax.scipy.stats as stats
from itertools import combinations
from typing import Callable
from copulax._src.univariate._utils import _univariate_input

from copulax._src.typing import Scalar
from copulax._src._utils import DEFAULT_RANDOM_KEY


class Correlation:
    r"""Class for computing correlation matrices."""

    # Making object a pytree
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        register_pytree_node(cls,
                             cls.tree_flatten,
                             cls.tree_unflatten)

    @classmethod
    def tree_unflatten(cls, aux_data, values, **init_kwargs):
        return cls(**init_kwargs)

    def tree_flatten(self):
        return (), None

    # Standard correlation matrix implementations
    def _insure_valid(self, A: Array) -> Array:
        lower_triangular: jnp.ndarray = jnp.tril(A)
        return jnp.fill_diagonal(lower_triangular + lower_triangular.T, 1.0, inplace=False)

    def pearson(self, x: jnp.ndarray) -> Array:
        r"""Pearson correlation matrix."""
        pearson: jnp.ndarray = jnp.corrcoef(x, rowvar=False)
        return self._insure_valid(pearson)


    def spearman(self, x: jnp.ndarray) -> Array:
        r"""Spearman-rank correlation matrix."""
        ranks: jnp.ndarray = stats.rankdata(x, axis=0)
        spearman: jnp.ndarray = self.pearson(ranks)
        return self._insure_valid(spearman)

    @jit
    def _kendall_pair_calc(x: jnp.ndarray, y: jnp.ndarray, 
                           i: int, j: int) -> Scalar:
        return jnp.where(i < j, 
                         jnp.sign(x[i] - x[j]) * jnp.sign(y[i] - y[j]), 
                         0.0)
    
    @staticmethod 
    @jit  
    def _kendall_iter(carry, j: int):
        """scans over i, s.t. i < j"""
        pair_sum, x, y, js = carry
        _iter: Callable = lambda ps, i: (
            ps + Correlation._kendall_pair_calc(x=x, y=y, i=i, j=j), None)
        pair_sum += lax.scan(_iter, 0.0, js)[0]
        return (pair_sum, x, y, js), None

    @staticmethod
    @jit
    def _kendall_pair(x: jnp.ndarray, y: jnp.ndarray, n: int, js: jnp.ndarray
                      ) -> Scalar:
        """scans over j"""
        scale: Scalar = 2 / (n * (n - 1))
        res: tuple = lax.scan(Correlation._kendall_iter, (0.0, x, y, js), js)
        return res[0][0] * scale

    @staticmethod
    @jit
    def _fill_kendall(carry: tuple, indices: tuple) -> Array:
        x, kendall, n, js = carry
        i, j = indices
        kendall_pair: Scalar = Correlation._kendall_pair(x[:, i], x[:, j], n, js)
        kendall = kendall.at[i, j].set(kendall_pair)
        kendall = kendall.at[j, i].set(kendall_pair)
        return (x, kendall, n, js), None
    
    def kendall(self, x: jnp.ndarray) -> Array:
        r"""Kendall-tau correlation matrix.
        
        Note:
            This implementation is essentially a brute force version of 
            the O(n^2) algorithm (O(d * n^2) when calculating the entire 
            correlation matrix) and hence is not efficient and should be
            avoided for large datasets.
        """
        n, d = x.shape
        js: jnp.ndarray = jnp.arange(0, n, 1)
        indices: tuple = jnp.array(tuple(combinations(range(d), 2)))
        kendall: jnp.ndarray = jnp.ones((d, d))
        kendall = lax.scan(self._fill_kendall, (x, kendall, n, js), indices)[0][1]
        return self._insure_valid(kendall)

    # Alternative correlation matrix implementations
    def pp_kendall(self, x: jnp.ndarray) -> Array:
        """Pseudo-Pearson Kendall correlation matrix.
        
        Note:
            This assumes that the data is elliptically distributed and 
            hence has no skewness. It does however provide a method of 
            estimating the correlation matrix when variances/covariances
            are undefined or infinate.
        """
        kendall: jnp.ndarray = self.kendall(x)
        pp_kendall: jnp.ndarray = jnp.sin(0.5 * jnp.pi * kendall)
        return self._insure_valid(pp_kendall)
    
    # Rousseeuw and Molenberghs's denoising technique
    def _rm_denoising(self, A: jnp.ndarray, delta) -> tuple:
        eigenvalues, eigenvectors = jnp.linalg.eigh(A)
        positive_eigenvalues = jnp.where(eigenvalues > 0.0, eigenvalues, delta)
        return positive_eigenvalues.real, eigenvectors.real
    
    def _rm_incomplete(self, A: jnp.ndarray, delta: Scalar) -> Array:
        positive_eigenvalues, eigenvectors = self._rm_denoising(A, delta)
        new_A: jnp.ndarray = (eigenvectors 
                              @ jnp.diag(positive_eigenvalues) 
                              @ jnp.linalg.inv(eigenvectors))
        return new_A

    def _rm(self, A: jnp.ndarray, delta: Scalar) -> Array:
        new_A: jnp.ndarray = self._rm_incomplete(A, delta)
        return self._insure_valid(new_A)

    def rm_pearson(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._rm(self.pearson(x), delta)
    
    def rm_spearman(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._rm(self.spearman(x), delta)

    def rm_kendall(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._rm(self.kendall(x), delta)
    
    def rm_pp_kendall(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._rm(self.pp_kendall(x), delta)
    
    # Laloux et al.'s denoising technique
    def _laloux(self, x: jnp.ndarray, A: jnp.ndarray, delta: Scalar) -> Array:
        # performing RM denoising
        positive_eigenvalues, eigenvectors = self._rm_denoising(A, delta)

        # calculating the Bulk
        n, d = x.shape
        Q: Scalar = n / d
        bulk_ub: Scalar = (1 + jnp.pow(Q, -0.5))**2
        
        # replacing eigenvalues with mean
        cond: jnp.ndarray = positive_eigenvalues > bulk_ub
        k: Scalar = jnp.sum(cond)
        denominator: Scalar = jnp.where(d-k > 0, d-k, 1.0)
        fill_val: Scalar = jnp.where(~cond, positive_eigenvalues, 0.0).sum() / denominator
        # fill_val: Scalar = jnp.where(k > 0, positive_eigenvalues[~cond].mean(), 0.0)
        new_eigenvalues: jnp.ndarray = jnp.where(cond, positive_eigenvalues, fill_val)

        # reconstructing the matrix
        laloux: jnp.ndarray = eigenvectors @ jnp.diag(new_eigenvalues) @ jnp.linalg.inv(eigenvectors)
        return self._insure_valid(laloux)
    
    def laloux_pearson(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._laloux(x, self.pearson(x), delta)
    
    def laloux_spearman(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._laloux(x, self.spearman(x), delta)
    
    def laloux_kendall(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._laloux(x, self.kendall(x), delta)
    
    def laloux_pp_kendall(self, x: jnp.ndarray, delta: Scalar = 1e-5) -> Array:
        return self._laloux(x, self.pp_kendall(x), delta)

    # def ledoit_wolf(self, x: jnp.ndarray) -> Array:
    #     pass

    # helper functions
    def _corr_from_cov(self, C: jnp.ndarray) -> Array:
        """Convert covariance matrix to correlation matrix."""
        # calculating the diagonal matrix of standard deviations
        sigma_diag: jnp.ndarray = jnp.diag(jnp.sqrt(jnp.diag(C)))
        diag_inv: jnp.ndarray = jnp.linalg.inv(sigma_diag)

        # returning the implied correlation matrix
        R: jnp.ndarray = diag_inv @ C @ diag_inv
        return R

    def _cov_from_vars(self, vars: jnp.ndarray, R: jnp.ndarray) -> Array:
        """Convert variances and correlation matrix to covariance matrix."""
        # calculating the diagonal matrix of standard deviations
        sigma_diag: jnp.ndarray = jnp.diag(jnp.sqrt(vars.flatten()))
        
        # returning the implied pseudo covariance matrix
        return sigma_diag @ R @ sigma_diag
    
    def _cov_from_corr(self, x: jnp.ndarray, R: jnp.ndarray) -> Array:
        """Convert correlation matrix to covariance matrix."""
        # calculating the variances of the input data
        vars: jnp.ndarray = jnp.var(x, axis=0)
        return self._cov_from_vars(vars=vars, R=R)


_corr: Correlation = Correlation()


def corr(x: ArrayLike, method: str = 'pearson', **kwargs) -> Array:
    r"""Compute the correlation matrix of the input data.

    Args:
        x: arraylike, input data.
        method: str, method to use for computing the correlation matrix. 
            Options are 'pearson', 'spearman', 'kendall', 'pp_kendall', 
            'rm_pearson', 'rm_kendall', 'rm_spearman', 'rm_pp_kendall', 
            'laloux_pearson', 'laloux_spearman', 'laloux_kendall' and 
            'laloux_pp_kendall'.

    Note:
        If you intend to jit wrap this function, ensure that 'method' 
        is a static argument.

    Returns:
        array, correlation matrix of the input data.
    """
    method: str = method.lower().strip()
    func: Callable = getattr(_corr, method, 'pearson')
    return func(x=x, **kwargs)


def cov(x: ArrayLike, method: str = 'pearson', **kwargs) -> Array:
    r"""Compute the covariance matrix of the input data. Note that many of the 
    available methods are correlation matrix-based and hence the 
    pseudo-covariance matrix (=covariance when method = 'pearson') 
    of sigma_diag * corr * sigma_diag is returned.

    Args:
        x: arraylike, input data.
        method: str, method to use for computing the correlation matrix. 
            Options are 'pearson', 'spearman', 'kendall', 'pp_kendall', 
            'rm_pearson', 'rm_kendall', 'rm_spearman', 'rm_pp_kendall', 
            'laloux_pearson', 'laloux_spearman', 'laloux_kendall' and 
            'laloux_pp_kendall'.

    Note:
        If you intend to jit wrap this function, ensure that 'method' 
        is a static argument.

    Returns:
        array, covariance matrix of the input data.
    """
    # calculating correlation matrix
    corr_matrix: jnp.ndarray = corr(x=x, method=method, **kwargs)

    # returning the implied pseudo covariance matrix
    return _corr._cov_from_corr(x=x, R=corr_matrix)


def random_correlation(size: int, key: Array = DEFAULT_RANDOM_KEY) -> Array:
    r"""Efficiently generates a random correlation matrix of given size.

    Note:
        If you intend to jit wrap this function, ensure that 'size' 
        is a static argument.

        Uses the factors method described in:
        https://stats.stackexchange.com/questions/124538/how-to-generate-a-large-full-rank-random-correlation-matrix-with-some-strong-cor

    Args:
        size (int): size of the correlation matrix.
        key (jax.random.PRNGKey): jax random key, used for generating 
            random numbers.
            
    Returns:
        rand_corr: random correlation matrix of given size.
    """
    # generating random covariance matrix
    key, subkey = random.split(key)
    W: Array = random.uniform(key=key, shape=(size, size), minval=-1.0, maxval=1.0)
    D: Array = jnp.diag(random.uniform(key=subkey, shape=(size,), minval=0.0, maxval=1.0))
    C: Array = W @ W.T + D
    
    # converting covariance matrix to correlation matrix
    R: Array = _corr._corr_from_cov(C=C)
    return R


def random_covariance(vars: Array, key: Array = DEFAULT_RANDOM_KEY) -> Array:
    r"""Efficiently generates a random covariance matrix of given size.

    Args:
        vars (Array): Variances of the covariates and implies the size 
            of the covariance matrix.
        key (jax.random.PRNGKey): jax random key, used for generating 
            random numbers.
    
    Returns:
        rand_cov: random covariance matrix of given size.
    """
    # we could simply use the same approach as in random_correlation, 
    # to generate the covariance matrix C. However, whilst this would
    # be more efficient and would negate the need for the vars argument,
    # the scale of the covariances in C can become large and disjoint 
    # from any relevant data distribution.
    vars, _ = _univariate_input(vars)
    R: Array = random_correlation(size=vars.size, key=key)
    return _corr._cov_from_vars(vars=vars, R=R)


"""File containing useful functions for mean-variance mixture distributions."""
from jax import lax


def mean_variance_ldmle_params(stats: dict, gamma: float, sample_mean: float, sample_variance: float) -> tuple[float, float]:
    # obtaining mu and sigma estimates
    mu: float = sample_mean - stats['mean'] * gamma
    sigma_sq: float = (sample_variance - stats['variance'] * lax.pow(gamma, 2)) / stats['mean'] 
    sigma: float = lax.sqrt(lax.abs(sigma_sq))
    return mu, sigma


def mean_variance_stats(w_stats: dict, mu: float, sigma: float, gamma: float) -> dict:
    # obtaining the mean and variance of the mixture distribution
    mean: float = mu + lax.mul(w_stats['mean'], gamma)
    var: float = lax.mul(w_stats['mean'], lax.pow(sigma, 2)) + lax.mul(w_stats['variance'], lax.pow(gamma, 2))
    std: float = lax.sqrt(var)
    return {'mean': mean, 'variance': var, 'std': std}


"""File containing the copuLAX implementaqtion of the Generalized Inverse 
Gaussian distribution."""
import jax.numpy as jnp
from jax import random, lax, custom_vjp, jit
from jax._src.typing import ArrayLike, Array
from copy import deepcopy

from copulax._src._distributions import Univariate
from copulax._src.typing import Scalar
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.univariate._cdf import _cdf, cdf_bwd, _cdf_fwd
from copulax._src.optimize import projected_gradient
from copulax.special import kv


class GIGBase(Univariate):
    @classmethod
    def _params_dict(cls, lamb: Scalar, chi: Scalar, psi: Scalar) -> dict:
        d: dict = {"lambda": lamb, "chi": chi, "psi": psi}
        return cls._args_transform(d)
    
    @staticmethod
    def _params_to_tuple(params: dict) -> tuple:
        params = GIGBase._args_transform(params)
        return params["lambda"], params["chi"], params["psi"]
    
    @staticmethod
    def _params_to_array(params: dict) -> Array:
        return jnp.asarray(GIGBase._params_to_tuple(params)).flatten()
    
    @classmethod
    def _support(cls, *args, **kwargs) -> tuple:
        return 0.0, jnp.inf
    
    def example_params(self, *args, **kwargs):
        r"""Example parameters for the GIG distribution.
        
        This is a three parameter family of continuous distributions, 
        with the GIG being defined by shape parameters `lambda`, `chi`, 
        and `psi`. Here, we adopt the parameterization used by McNeil 
        et al. (2005)"""
        return self._params_dict(lamb=1.0, chi=1.0, psi=1.0)
    
    @staticmethod
    def _stable_logpdf(stability: Scalar, x: ArrayLike, params: dict) -> Array:
        lamb, chi, psi = GIGBase._params_to_tuple(params)
        x, xshape = _univariate_input(x)

        var = lax.add(lax.mul(lamb - 1, lax.log(x)), 
                    -0.5 * (lax.mul(chi, lax.pow(x, -1)) + lax.mul(psi, x)))

        cT = lax.mul(0.5*lamb, lax.log((psi/chi) + stability))
        kv_val = kv(lamb, lax.pow(lax.mul(chi, psi), 0.5))
        cB = lax.log(stability + 2 * kv_val)
        
        c = lax.sub(cT, cB)
        logpdf_raw = lax.add(var, c)
        logpdf: jnp.ndarray = jnp.where(jnp.isnan(logpdf_raw), -jnp.inf, logpdf_raw)
        return logpdf.reshape(xshape)
    
    @staticmethod
    def logpdf(x: ArrayLike, params: dict) -> Array:
        return GIGBase._stable_logpdf(stability=0.0, x=x, params=params)
    
    @staticmethod
    def pdf(x: ArrayLike, params: dict) -> Array:
        return lax.exp(GIGBase.logpdf(x=x, params=params))
    
    # ppf
    # def _get_x0(self, params: dict) -> Scalar:
    #     return self.stats(params=params)['mode']

    # sampling
    # Uses the method outlined by Luc Devroye in "Random variate generation for 
    # the generalized inverse Gaussian distribution" (2014).
    def _devroye(self, x, alpha, lamb):
        return -alpha * (jnp.cosh(x) - 1) - lamb * (jnp.exp(x) - x - 1)

    def _devroye_grad(self, x, alpha, lamb):
        return -alpha * jnp.sinh(x) - lamb * (jnp.exp(x) - 1)
    
    def _new_single_rv(self, carry, _):
        key, _, stop, count, constants = carry
        lamb, alpha, t, s, t_, s_, eta, zeta, theta, xi, p, r, q = constants


        key, subkey = random.split(key)
        u, v, w = random.uniform(subkey, shape=(3, ))

        x = jnp.where(u < (q + r) / (q + p + r), t_ + r * lax.log(1 / v), -s_ - p * lax.log(1 / v ))
        x = jnp.where(u < q / (q + p + r), -s_ + q * v, x)

        # checking stopping condition
        chi = jnp.where(jnp.logical_and(-s_ < x, x < t_), 1.0, 0.0) + jnp.where(t_ < x, jnp.exp(-eta - zeta * (x - t)) , 0.0) + jnp.where(x < -s_, jnp.exp(-theta + xi * (x + s)), 0.0)
        stop = w * chi <= jnp.exp(self._devroye(x, alpha, lamb))

        return (key, x, stop, count + 1, constants), None
    
    @jit
    def _generate_single_rv(self, key: Array, constants: tuple) -> tuple[Array, Array]:
        maxiter = 10
        init = (key, jnp.array(jnp.nan), False, 0, constants)
        res = lax.scan((lambda carry, _: lax.cond(carry[2], (lambda carry, _: (carry, _)), self._new_single_rv, carry, None)), init, None, maxiter)[0]
        return res[0], res[1]
    
    def rvs(self, size: tuple | Scalar, params: dict, key: Array=DEFAULT_RANDOM_KEY) -> Array:
        # getting parameters
        lamb, chi, psi = self._params_to_tuple(params)
        sign_lamb: int = jnp.where(jnp.sign(lamb) >= 0, 1, -1)
        lamb: float = jnp.abs(lamb)
        omega: float = lax.sqrt(chi * psi)
        alpha: float = lax.sqrt(jnp.pow(omega, 2) + jnp.pow(lamb, 2)) - lamb

        # getting positive constant t
        _devroye_1: float = self._devroye(x=1, alpha=alpha, lamb=lamb)
        t: float = jnp.where(-_devroye_1 > 2, lax.sqrt(2 / (alpha + lamb)), 1)
        t = jnp.where(-_devroye_1 < 0.5, lax.log(4/(alpha + 2*lamb)), t)
        
        # getting positive constant s
        _devroye_minus_1: float = self._devroye(x=-1, alpha=alpha, lamb=lamb)
        s: float = jnp.where(-_devroye_minus_1 > 2, lax.sqrt(4 / (alpha * jnp.cosh(1) + lamb)), 1)
        s = jnp.where(-_devroye_minus_1 < 0.5, jnp.min(jnp.array([1 / lamb, lax.log(1 + (1 / alpha) + lax.sqrt(jnp.pow(alpha, -2) + (2 / alpha))),]))  , s)
        
        # Computing constants
        eta, zeta, theta, xi = -self._devroye(x=t, alpha=alpha, lamb=lamb), -self._devroye_grad(x=t, alpha=alpha, lamb=lamb), -self._devroye(x=-s, alpha=alpha, lamb=lamb), self._devroye_grad(x=-s, alpha=alpha, lamb=lamb)
        p, r = 1 / xi, 1 / zeta
        t_: float = t - r * eta
        s_: float = s - p * theta
        q : float = t_ + s_

        # Generating random variables
        constants: tuple = (lamb, alpha, t, s, t_, s_, eta, zeta, theta, xi, p, r, q)
        if isinstance(size, (int, float)):
            num_samples: int = int(size)
        else:
            num_samples: int = 1
            for number in size:
                num_samples *= number

        X: jnp.ndarray = lax.scan((lambda key, _ : self._generate_single_rv(key, constants)), key, None, num_samples)[1]

        frac: float = lax.div(lamb, omega)
        c: float = frac + lax.sqrt(1 + lax.pow(frac, 2))

        return jnp.pow((c * jnp.exp(X)), sign_lamb).reshape(size)
    
    # stats
    def _sample_estimates(self, params: dict, analytical_mean: Scalar, analytical_variance: Scalar) -> tuple[Scalar, Scalar]:
        sample = self.rvs(size=1000, params=params)
        sample_mean = jnp.mean(sample)
        sample_variance = jnp.var(sample)
        return (sample_mean, sample_variance)

    def stats(self, params: dict) -> dict:
        lamb, chi, psi = self._params_to_tuple(params)

        # calculating mean
        r: float = lax.sqrt(lax.mul(chi, psi))
        frac: float = lax.div(chi, psi)
        kv_lamb: float = kv(lamb, r)
        kv_lamb_plus_1: float = kv(lamb + 1, r)
        analytical_mean: float = lax.mul(lax.pow(frac, 0.5), lax.div(kv_lamb_plus_1, kv_lamb))

        # calculating variance
        kv_lamb_plus_2: float = kv(lamb + 2, r)
        second_moment: float = lax.mul(frac, lax.div(kv_lamb_plus_2, kv_lamb))
        analytical_variance: float = lax.sub(second_moment, lax.pow(analytical_mean, 2))

        # accounting for numerical instability
        # when psi is very large, the first and second moments can be 
        # NaN due to divide by zero error. Resolving this by using sample
        # mean and variance estimates in these cases.
        mean, variance = lax.cond(
            jnp.logical_or(jnp.isnan(analytical_mean), jnp.isnan(analytical_variance)), 
            self._sample_estimates, 
            (lambda params, analytical_mean, analytical_variance: (analytical_mean, analytical_variance)), params, analytical_mean, analytical_variance)
        
        std: float = lax.sqrt(variance)

        # mode
        mode: float = lax.div((lamb - 1) + lax.sqrt(lax.pow(lamb - 1, 2) + lax.mul(chi, psi)), psi)
        
        return self._scalar_transform({
            'mean': mean, 
            'variance': variance, 
            'std': std,
            'mode': mode})
    
    # fitting
    def _fit_mle(self, x: jnp.ndarray, lr: float, maxiter: int) -> dict:
        eps = 1e-8
        constraints: tuple = (jnp.array([[-jnp.inf, eps, eps]]).T, 
                            jnp.array([[jnp.inf, jnp.inf, jnp.inf]]).T)
        
        projection_options: dict = {'hyperparams': constraints}

        key1, key = random.split(DEFAULT_RANDOM_KEY)
        key2, key3 = random.split(key)
        params0: jnp.ndarray = jnp.array([
            random.normal(key1, ()), 
            random.uniform(key2, (), minval=eps), 
            random.uniform(key3, (), minval=eps)])
        
        res = projected_gradient(
            f=self._mle_objective, x0=params0, projection_method='projection_box', 
            projection_options=projection_options, x=x, lr=lr, maxiter=maxiter)
        lamb, chi, psi = res['x']
        return self._params_dict(lamb=lamb, chi=chi, psi=psi)#, res['fun']
    
    def fit(self, x: ArrayLike, lr: float = 0.1, maxiter: int = 100) -> dict:
        r"""Fit the distribution to the input data.
        
        Args:
            x (ArrayLike): The input data to fit the distribution to.
            lr (float): Learning rate for optimization.
            maxiter (int): Maximum number of iterations for optimization.
        
        Returns:
            dict: The fitted distribution parameters.
        """ 
        x: jnp.ndarray = _univariate_input(x)[0]
        return self._fit_mle(x=x, lr=lr, maxiter=maxiter)
    
    # cdf
    @staticmethod
    def _params_from_array(params_arr, *args, **kwargs) -> dict:
        lamb, chi, psi = params_arr
        return GIGBase._params_dict(lamb=lamb, chi=chi, psi=psi)

    @staticmethod
    def _pdf_for_cdf(x: ArrayLike, *params_tuple) -> Array:
        params_array: jnp.ndarray = jnp.asarray(params_tuple).flatten()
        params: dict = GIGBase._params_from_array(params_array)
        return GIG.pdf(x=x, params=params)
    

def _vjp_cdf(x: ArrayLike, params: dict) -> Array:
    params: dict = GIGBase._args_transform(params)
    return _cdf(dist=GIGBase, x=x, params=params)


_vjp_cdf_copy = deepcopy(_vjp_cdf)
_vjp_cdf = custom_vjp(_vjp_cdf)


def cdf_fwd(x: ArrayLike, params: dict) -> tuple[Array, tuple]:
    params = GIGBase._args_transform(params)
    return _cdf_fwd(dist=GIGBase, cdf_func=_vjp_cdf_copy, x=x, params=params)


_vjp_cdf.defvjp(cdf_fwd, cdf_bwd)


class GIG(GIGBase):
    r"""The Generalized Inverse Gaussian distribution is a 3 parameter family 
    of continuous distributions.
    
    We adopt the parameterization used by McNeil et al. (2005)

    https://en.wikipedia.org/wiki/Generalized_inverse_Gaussian_distribution

    :math: `\lambda` is real-valued.
    :math: `\chi` is strictly positive.
    :math: `\psi` is strictly positive.
    """
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        return _vjp_cdf(x=x, params=params)


gig = GIG("GIG")


import pytest
import numpy as np


NUM_SAMPLES: int = 100
NUM_ASSETS: int = 3


np.random.seed(0)


def gen_uncorrelated_data(num_samples: int, num_assets: int):
    return np.random.normal(size=(num_samples, num_assets))


def gen_correlated_data(num_assets: int, num_samples: int):
    # generating a random correlation matrix
    random_uniform: np.ndarray = np.random.uniform(-1, 1, size=(num_assets, num_assets))
    lower_triangular: np.ndarray = np.tril(random_uniform)
    correlation: np.ndarray = lower_triangular + lower_triangular.T
    np.fill_diagonal(correlation, 1.0)

    # converting to covariance matrix
    sigma_diag: np.ndarray = np.diag(np.random.uniform(1, 4, num_assets))
    covariance_matrix: np.ndarray = sigma_diag @ correlation @ sigma_diag

    # generating correlated data
    return np.random.multivariate_normal(np.zeros(num_assets), covariance_matrix, num_samples)


@pytest.fixture(scope='package', autouse=True)
def uncorrelated_sample():
    return gen_uncorrelated_data(NUM_SAMPLES, NUM_ASSETS)


@pytest.fixture(scope='package', autouse=True)
def correlated_sample():
    return gen_correlated_data(NUM_ASSETS, NUM_SAMPLES)


@pytest.fixture(scope='package', autouse=True)
def uncorrelated_small_sample():
    return gen_uncorrelated_data(NUM_ASSETS, NUM_ASSETS)


@pytest.fixture(scope='package', autouse=True)
def correlated_small_sample():
    return gen_correlated_data(NUM_ASSETS+1, NUM_ASSETS)

@pytest.fixture(scope='package', autouse=True)
def datasets(uncorrelated_sample, correlated_sample, uncorrelated_small_sample, correlated_small_sample):
    return {'uncorrelated_sample': gen_uncorrelated_data(NUM_SAMPLES, NUM_ASSETS),
            'correlated_sample': gen_correlated_data(NUM_ASSETS, NUM_SAMPLES), 
            
            # 'too_small_dim_sample': gen_uncorrelated_data(NUM_SAMPLES, 1),
            'too_large_dim_sample': gen_uncorrelated_data(NUM_SAMPLES, 100),
            }





"""Tests for univariate probability distributions."""
import numpy as np
from jax import jit, grad
import inspect
import pytest
from jax import numpy as jnp

from copulax._src._distributions import Univariate
from copulax.univariate import distributions
from copulax.univariate.distributions import _all_dists, skewed_t
from copulax._src.typing import Scalar
from copulax.tests.helpers import *


# Helper functions for testing univariate distributions
def get_data(dist, continuous_data, discrete_data):
    """Get data for testing."""
    if dist.dtype == 'continuous':
        return continuous_data
    elif dist.dtype == 'discrete':
        return discrete_data
    else:
        raise ValueError(f"Unknown distribution type: {dist.dtype}")


# Test combinations
DISTRIBUTIONS = tuple(value for value in distributions.values() if isinstance(value, Univariate)) # note has to be a tuple otherwise pytest parametrize will remove elements as it scans along, causing test skipping
IVTF_DISTRIBUTIONS = tuple()
UNNORMALISED_DISTRIBUTIONS = tuple((skewed_t.name,))
rvs_sizes = (
        ((), 1),
        ((1, ), 1),
        ((1, 1), 1),
        ((3, ), 3),
        ((3, 1), 3),
        ((3, 2), 6),
        (0, 0),
        (1, 1),
        (3, 3),
        )
RVS_COMBINATIONS = tuple((dist, size) for dist in DISTRIBUTIONS for size in rvs_sizes)
IVTF_RVS_COMBINATIONS = tuple((dist, size) for dist in IVTF_DISTRIBUTIONS for size in rvs_sizes)


# Tests
@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_all_objects(dist):
    assert isinstance(dist, Univariate), f"{dist} is not a Univariate object."


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_all_methods_implemented(dist):
    methods: set[str] = {
        'support', 'logpdf', 'pdf', 'logcdf', 'cdf', 'ppf', 'inverse_cdf', 
        'rvs', 'sample', 'fit', 'stats', 'loglikelihood', 'aic', 'bic', 'dtype', 
        'dist_type', 'name', 'example_params'
        }
    
    for method_name in methods:
        assert hasattr(dist, method_name), f"{dist} missing {method_name} method."

    # testing no additional methods are implemented
    pytree_methods: set[str] = {'tree_flatten', 'tree_unflatten'}
    extra_methods = set(dist.__dict__.keys()) - methods - pytree_methods
    extra_methods = {m for m in extra_methods if not m.startswith('_')}
    assert not extra_methods, f"{dist} has extra methods: {extra_methods}"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_name(dist):
    assert isinstance(dist.name, str), f"{dist} name is not a string"
    assert dist.name != "", f"{dist} name is an empty string"
    assert dist.name == str(dist), f"{dist} name does not match its string representation"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dtype(dist):
    assert isinstance(dist.dtype, str), f"{dist} dtype is not a string"
    assert dist.dtype != "", f"{dist} dtype is an empty string"
    assert dist.dtype in ('continuous', 'discrete'), f"dtype is not 'continuous' or 'discrete' for {dist}"
    

@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dist_type(dist):
    assert isinstance(dist.dist_type, str), f"{dist} dist_type is not a string"
    assert dist.dist_type != "", f"{dist} dist_type is an empty string"
    assert dist.dist_type == 'univariate', f"dist_type is not 'univariate' for {dist}"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_example_params(dist):
    params = dist.example_params()

    # testing properties
    s = f'{dist} example_params'
    check_uvt_params(params, s)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dist_object_jittable(dist):
    jittable(dist)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_support(dist):
    params: dict = dist.example_params()
    support = dist.support(params)
    
    # testing properties
    assert isinstance(support, jnp.ndarray), f"{dist} support is not a JAX array"
    assert support.size == 2, f"{dist} does not contain 2 elements"
    assert support.ndim == 1 and support.shape == (2,), f"{dist} support is not a flattened array"
    assert np.all(support[0] < support[1]), f"{dist} support bounds are not in order."
    assert no_nans(support), f"{dist} support contains NaNs."
    
    # testing jit works
    jitted_support = jit(dist.support)(params)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_logpdf(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)
    params: dict = dist.example_params()

    output = dist.logpdf(data, params)
    
    # testing properties
    correct_uvt_shape(data, output, dist, "logpdf")
    assert no_nans(output), f"{dist.name} logpdf contains NaNs."
    
    # testing jit works
    jitted_pdf = jit(dist.logpdf)(data, params)
    
    # testing gradients
    gradients(dist.logpdf, f"{dist.name} logpdf", data, params, params_error=dist.name not in UNNORMALISED_DISTRIBUTIONS)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_pdf(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)
    params: dict = dist.example_params()

    output = dist.pdf(data, params)

    # Check properties
    correct_uvt_shape(data, output, dist, "pdf")
    assert no_nans(output), f"{dist} pdf contains NaNs."
    assert is_positive(output), f"{dist} pdf contains negative values."
    assert is_finite(output), f"{dist} pdf contains non-finite values."

    # testing jit works
    jitted_pdf = jit(dist.pdf)(data, params)
    
    # testing gradients
    gradients(dist.pdf, f"{dist.name} pdf", data, params, params_error=dist.name not in UNNORMALISED_DISTRIBUTIONS)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_logcdf(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)
    params: dict = dist.example_params()

    output = dist.logcdf(data, params)
    
    # testing properties
    correct_uvt_shape(data, output, dist, "logcdf")
    assert no_nans(output), f"{dist} logcdf contains NaNs."
    
    # testing jit works
    jitted_logcdf = jit(dist.logcdf)(data, params)
    
    # testing gradients
    gradients(dist.logcdf, f"{dist.name} logcdf", data, params, params_error=dist.name not in UNNORMALISED_DISTRIBUTIONS)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_cdf(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)
    params: dict = dist.example_params()

    output = dist.cdf(data, params)

    # Check properties
    correct_uvt_shape(data, output, dist, "cdf")
    assert no_nans(output), f"{dist} cdf contains NaNs."
    assert is_positive(output), f"{dist} cdf contains negative values."
    assert is_finite(output), f"{dist} cdf contains non-finite values."
    assert np.all(0 <= output) and np.all(output <= 1), f"cdf not in [0, 1] range for {dist}"
    
    # testing jit works
    jitted_cdf = jit(dist.cdf)(data, params)
    
    # testing gradients
    gradients(dist.cdf, f"{dist.name} cdf", data, params, params_error=dist.name not in UNNORMALISED_DISTRIBUTIONS)


def _ppf(dist, uniform_data, cubic):
    """Test the ppf function."""
    data: jnp.ndarray = uniform_data
    params: dict = dist.example_params()
    support = dist.support(params)
    s: str = " with cubic spline approximation" if cubic else ""

    output = dist.ppf(data, params, cubic=cubic, maxiter=5)
    
    # testing properties
    correct_uvt_shape(data, output, dist, "ppf")
    assert no_nans(output), f"{dist} ppf contains NaNs{s}."
    assert is_finite(output), f"{dist} ppf contains non-finite values{s}."
    assert np.all(output >= support[0]) and np.all(output <= support[1]), f"ppf not in support range for {dist}{s}."

    # testing jit works
    jitted_ppf_func = jit(dist.ppf, static_argnames='cubic')
    jitted_ppf = jitted_ppf_func(data, params, cubic=cubic)

    # testing gradients
    gradients(jitted_ppf_func, f"{dist.name} ppf", data, params, cubic=cubic)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_ppf(dist, uniform_data):
    _ppf(dist, uniform_data, cubic=False)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_ppf_cubic(dist, uniform_data):
    _ppf(dist, uniform_data, cubic=True)


def _rvs(dist, tup):
    """Test the rvs function."""
    params: dict = dist.example_params()
    size, num = tup

    output = dist.rvs(size, params)

    # testing properties
    if isinstance(size, int):
        shape = (size, )
    else:
        shape = size
    assert isinstance(output, jnp.ndarray), f"{dist} rvs is not a JAX array."
    assert no_nans(output), f"{dist} rvs contains NaNs."
    assert is_finite(output), f"{dist} rvs contains non-finite values."
    assert output.size == num, f"{dist} rvs size mismatch."
    assert output.shape == shape, f"{dist} rvs shape mismatch."

    # testing jit works
    jitted_rvs = jit(dist.rvs, static_argnames='size')(size=size, params=params)


@pytest.mark.parametrize("dist, size", RVS_COMBINATIONS)
def test_rvs(dist, size):
    _rvs(dist, size)


@pytest.mark.local_only
@pytest.mark.parametrize("dist, size", IVTF_RVS_COMBINATIONS)
def test_rvs_ivtf(dist, size):
    _rvs(dist, size)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_fit(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)

    fitted_params = dist.fit(data)

    # testing properties
    check_uvt_params(fitted_params, f"{dist} fitted")
    assert set(fitted_params.keys()) == set(dist.example_params().keys()), f"{dist} fitted params and example_params mismatch."
    
    # testing jit works
    fit_args: list = inspect.getfullargspec(dist.fit).args
    if 'method' in fit_args:
        jit_fit = jit(dist.fit, static_argnames='method')(continuous_data)
    else:
        jit_fit = jit(dist.fit)(continuous_data)
    check_uvt_params(fitted_params, f"{dist} jit fitted")


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_stats(dist):
    params: dict = dist.example_params()
    stats = dist.stats(params=params)

    # testing properties    
    assert isinstance(stats, dict), f"{dist} stats outputted wrong type"
    assert len(stats) > 0, f"{dist} stats is empty"
    assert all(isinstance(k, str) for k in stats.keys()), f"{dist} stats keys are not strings"
    assert all(isinstance(v, Scalar) for v in stats.values()), f"{dist} stats values are not scalars"
    assert all(v.ndim == 0 and v.shape == () and v.size == 1 for v in stats.values()), f"{dist} stats values are not scalars"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_loglikelihood(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)
    params: dict = dist.example_params()

    output = dist.loglikelihood(data, params)

    # testing properties
    check_metric_output(dist, output, "loglikelihood")

    # testing jit works
    jitted_loglikelihood = jit(dist.loglikelihood)(data, params)

    # testing gradients
    gradients(dist.loglikelihood, f"{dist} loglikelihood", data, params)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_aic(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)
    params: dict = dist.example_params()

    output = dist.aic(data, params)

    # testing properties
    check_metric_output(dist, output, "aic")

    # testing jit works
    jitted_aic = jit(dist.aic)(data, params)

    # testing gradients
    gradients(dist.aic, f"{dist} aic", data, params)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_bic(dist, continuous_data, discrete_data):
    data: jnp.ndarray = get_data(dist, continuous_data, discrete_data)
    params: dict = dist.example_params()

    output = dist.bic(data, params)

    # testing properties
    check_metric_output(dist, output, "bic")

    # testing jit works
    jitted_bic = jit(dist.bic)(data, params)

    # testing gradients
    gradients(dist.bic, f"{dist} bic", data, params)



"""Distribution fit metrics."""
import jax.numpy as jnp
from typing import Callable

from copulax._src.metrics import (
    _loglikelihood, 
    _mle_objective,
      _aic as __aic, 
      _bic as __bic,
      )


def _aic(logpdf_func: Callable, params: dict, x: jnp.ndarray, **kwargs) -> float:
    """Akaike Information Criterion (AIC) for model selection.
    Best model selected via minimising AIC.

    Args:
        logpdf_func: The log probability density function of the distribution.
        params: The parameters of the distribution.
        x: The data.
    """
    return __aic(logpdf_func=logpdf_func, params=params, x=x, k=len(params), **kwargs)


def _bic(logpdf_func: Callable, params: dict, x: jnp.ndarray, **kwargs) -> float:
    """Bayesian Information Criterion (BIC) for model selection.
    Best model selected via minimising BIC.

    Args:
        logpdf_func: The log probability density function of the distribution.
        params: The parameters of the distribution.
        x: The data.
    """
    return __bic(logpdf_func=logpdf_func, params=params, x=x, k=len(params), **kwargs)


"""File containing utility functions for multivariate distributions."""
import jax.numpy as jnp
from jax._src.typing import ArrayLike


def _multivariate_input(x: ArrayLike, dtype=float) -> tuple[jnp.ndarray, tuple[int], int]:
    """Ensures all input arrays are of the same dtype and (n, d) shape."""
    x_arr: jnp.array = jnp.asarray(x, dtype=dtype)
    xshape = x_arr.shape
    n, d = xshape
    return x_arr.reshape((n, d)), (n, 1), n, d


"""Tests for multivariate probability distributions."""
import pytest
import jax.numpy as jnp
from jax import grad, jit
import numpy as np

from copulax._src._distributions import Multivariate
from copulax.tests.helpers import *
from copulax._src.typing import Scalar
from copulax.multivariate import mvt_normal, mvt_student_t, mvt_gh
from copulax.tests.multivariate.conftest import NUM_ASSETS

# Test combinations
DISTRIBUTIONS = tuple((mvt_normal, mvt_student_t, mvt_gh))
ERROR_CASES = tuple(('too_large_dim_sample',))
DATASETS = tuple(('uncorrelated_sample', 'correlated_sample', *ERROR_CASES))
COMBINATIONS = tuple((dist, dataset) for dist in DISTRIBUTIONS for dataset in DATASETS)


# Tests for multivariate distributions
@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_all_objects(dist):
    assert isinstance(dist, Multivariate), f"{dist} is not a Multivariate object"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_all_methods_implemented(dist):
    methods: set[str] = {
        'support', 'logpdf', 'pdf', 'rvs', 'sample', 'fit', 'stats', 
        'loglikelihood', 'aic', 'bic', 'dtype', 'dist_type', 'name', 
        'example_params',
    }
    # testing desired methods are implemented
    for method in methods:
        assert hasattr(dist, method), f"{dist} missing method {method}"

    # testing no additional methods are implemented
    pytree_methods: set[str] = {'tree_flatten', 'tree_unflatten'}
    extra_methods = set(dist.__dict__.keys()) - methods - pytree_methods
    extra_methods = {m for m in extra_methods if not m.startswith('_')}
    assert not extra_methods, f"{dist} has extra methods: {extra_methods}"
    

@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_name(dist):
    assert isinstance(dist.name, str), f"{dist} name is not a string"
    assert dist.name != "", f"{dist} name is an empty string"
    assert dist.name == str(dist), f"{dist} name does not match its string representation"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dtype(dist):
    assert isinstance(dist.dtype, str), f"{dist} dtype is not a string"
    assert dist.dtype != "", f"{dist} dtype is an empty string"
    assert dist.dtype in ('continuous', 'discrete'), f"dtype is not 'continuous' or 'discrete' for {dist}"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dist_type(dist):
    assert isinstance(dist.dist_type, str), f"{dist} dist_type is not a string"
    assert dist.dist_type != "", f"{dist} dist_type is an empty string"
    assert dist.dist_type == 'multivariate', f"dist_type is not 'multivariate' for {dist}"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_example_params(dist):
    params = dist.example_params()

    # checking properties
    s = f"{dist} example_params"
    check_mvt_params(params, s)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dist_object_jitable(dist):
    jittable(dist)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_support(dist):
    params: dict = dist.example_params()
    support = dist.support(params)
    
    # Checking properties
    assert isinstance(support, jnp.ndarray), f"{dist} support is not a JAX array"
    assert support.ndim == 2, f"{dist} support is not a 2D array"
    assert np.all(support[:, 0] < support[:, 1]), f"{dist} support bounds are not in order."
    assert no_nans(support), f"{dist} support contains NaNs."
    
    # Check jit
    jitted_support = jit(dist.support)(params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_logpdf(dist, dataset, datasets):
    data = datasets[dataset]
    params = dist.example_params()

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.logpdf(data, params=params)
    else:
        # Check non-error cases
        output = dist.logpdf(data, params=params)
        # Check properties
        correct_mvt_shape(data, output, dist, 'logpdf')
        assert no_nans(output), f"{dist} logpdf contains NaNs."
        # Check jit
        jitted_pdf = jit(dist.logpdf)(data, params=params)
        # Check gradients
        gradients(dist.logpdf, f"{dist} logpdf", data, params=params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_pdf(dist, dataset, datasets):
    data = datasets[dataset]
    params = dist.example_params()

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.pdf(data, params=params)
    else:
        # Check non-error cases
        output = dist.pdf(data, params=params)
        # Check properties
        correct_mvt_shape(data, output, dist, 'pdf')
        assert is_positive(output), f"{dist} pdf contains negative values."
        assert no_nans(output), f"{dist} pdf contains NaNs."
        assert is_finite(output), f"{dist} pdf contains non-finite values."
        # Check jit
        jitted_pdf = jit(dist.pdf)(data, params=params)
        # Check gradients
        gradients(dist.pdf, f"{dist} pdf", data, params)


SIZES = (0, 1, 2, 11)
SIZE_COMBINATIONS = tuple((dist, size) for dist in DISTRIBUTIONS for size in SIZES)
@pytest.mark.parametrize("dist, size", SIZE_COMBINATIONS)
def test_rvs(dist, size):
    params: dict = dist.example_params()
    output = dist.rvs(size=size, params=params)
    # Check properties
    expected_shape = (size, NUM_ASSETS)
    assert output.shape == expected_shape, f"{dist} rvs has incorrect shape. Expected {expected_shape}, got {output.shape}"
    assert no_nans(output), f"{dist} rvs contains NaNs for size {size}"
    assert is_finite(output), f"{dist} rvs contains infinite values for size {size}"
    # Check jit
    jitted_rvs = jit(dist.rvs, static_argnums=0)(size=size, params=params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_fit(dist, dataset, datasets):
    if dataset not in ERROR_CASES:
        data = datasets[dataset]
        params = dist.fit(data)
        # Check properties
        check_mvt_params(params, f"{dist} fit")
        # Check jit
        jitted_fit = jit(dist.fit)(data)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def stats(dist):
    params: dict = dist.example_params()
    stats = dist.stats(params=params)
    # Check properties
    assert isinstance(stats, dict), f"{dist} stats outputted wrong type"
    assert len(stats) > 0, f"{dist} stats is empty"
    assert all(isinstance(k, str) for k in stats.keys()), f"{dist} stats keys are not strings"
    assert all(isinstance(v, jnp.ndarray) for v in stats.values()), f"{dist} stats values are not arrays"
    assert all((v.ndim == 0 and v.shape == () and v.size == 1) or (v.ndim == 1 and v.shape == (v.size, 1) and v.size > 0) or (v.ndim == 2 and v.shape == (int(v.size ** 0.5), int(v.size ** 0.5)) and v.size > 0) for v in stats.values()), f"{dist} stats values are not scalars, 1D-vectors or 2D-square matrices"
    # Check jit
    jitted_stats = jit(dist.stats)(params=params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_loglikelihood(dist, dataset, datasets):
    data = datasets[dataset]
    params = dist.example_params()

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.loglikelihood(data, params=params)
    else:
        # Check non-error cases
        output = dist.loglikelihood(data, params=params)

        # Check properties
        check_metric_output(dist, output, 'loglikelihood')

        # Check jit
        jitted_loglikelihood = jit(dist.loglikelihood)(data, params=params)

        # Check gradients
        gradients(dist.loglikelihood, f"{dist} loglikelihood", data, params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_aic(dist, dataset, datasets):
    data = datasets[dataset]
    params = dist.example_params()

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.aic(data, params=params)
    else:
        # Check non-error cases
        output = dist.aic(data, params=params)

        # Check properties
        check_metric_output(dist, output, 'aic')

        # Check jit
        jitted_aic = jit(dist.aic)(data, params=params)

        # Check gradients
        gradients(dist.aic, f"{dist} aic on {dataset}", data, params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_bic(dist, dataset, datasets):
    data = datasets[dataset]
    params = dist.example_params()

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.bic(data, params=params)
    else:
        # Check non-error cases
        output = dist.bic(data, params=params)

        # Check properties
        check_metric_output(dist, output, 'bic')

        # Check jit
        jitted_bic = jit(dist.bic)(data, params=params)
        
        # Check gradients
        gradients(dist.bic, f"{dist} bic", data, params)


import jax.numpy as jnp
import jax
from typing import Callable
import jaxopt.projection as proj
from functools import partial
from jax._src.typing import Array

from copulax._src.typing import Scalar


###############################################################################
# ADAM optimizer
###############################################################################
@jax.jit
def adam(grad: jnp.ndarray, m: jnp.ndarray, v: jnp.ndarray, t: int, 
         beta1: float = 0.9, beta2: float = 0.999, eps: float = 1e-8,
         ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, int]:
    """Adam optimiser.

    .. math::
        https://arxiv.org/abs/1412.6980

    Args:
    grad: the gradient at the current iteration.
    m: the first moment estimate vector from the prior iteration.
    v: the second raw moment estimate vector from the prior iteration.
    t: the prior iteration count.
    beta1: the beta1 parameter controling exponential decay rate for the m vector.
    beta2: the beta2 parameter controling exponential decay rate for the v vector.
    eps: the epsilon parameter to prevent division by zero. Defaults to 1e-8.

    Returns:
        Adam direction, the first moment estimate vector, the second moment 
        estimate vector, the current iteration.
    """
    t += 1
    m = beta1 * m + (1 - beta1) * grad
    v = beta2 * v + (1 - beta2) * grad**2
    m_hat = m / (1 - beta1**(t + 1))
    v_hat = v / (1 - beta2**(t + 1))
    d = m_hat / (jnp.sqrt(v_hat) + eps)
    return d, m, v, t


@partial(jax.jit, static_argnames=("projection",))
def single_update(x: jnp.ndarray, d: jnp.ndarray, lr: float,
                  projection: Callable, projection_options: dict,) -> jnp.ndarray:
    """Update the weights using the projected gradient method.

    Args:
    x: the current parameters from the previous iteration.
    d: the direction to move in in the non-constrained space.
    lr: learning rate used in projected gradient descent.
    projection: the projection function to use.
    projection_options: dictionary of options for the projection function.

    Returns:
        The updated weights.
    """
    # Calculate the new weights
    x_uc: jnp.ndarray = x - lr * d

    # Project the new weights onto the feasible set
    x_uc = x_uc[None].T
    x_proj: jnp.ndarray = projection(x_uc, **projection_options)
    return x_proj.flatten()


def projected_gradient(f: Callable, x0: jnp.ndarray, projection_method: str,
                       lr: float = 1.0, maxiter: int = 100, 
                       adam_options: dict = {}, jit_options: dict = {}, 
                       projection_options: dict = {}, **kwargs
                       ) -> dict:
    """Projected gradient descent for linearly constrained optimization.
    Ninimizes the objective function f using the projected gradient descent algorithm and Adam gradient updates.

    Args:
        f: objective function to minimize. Must be jax.grad and jax.jit compatible and return a scalar value. The first argument must be the parameter vector to optimize.
        x0: initial guess. Must be a flatterned array with the same size as the solution.
        projection_method: name of the projection function to use. All jaxopt constrained optimisation projection functions are supported.
        lr: learning rate used in projected gradient descent.
        maxiter: maximum number of iterations.
        adam_options: dictionary of options for the Adam optimizer. 
        jit_options: kwargs to pass to jax.jit when compiling f.
        projection_options: kwargs to pass to the specified projection function.
        kwargs: additional arguments to pass to the objective function.

    Returns:
        dictionary containing optimal results.
    """
    # JIT compiling the projection and gradient functions
    projection: Callable = getattr(proj, projection_method)
    projection = jax.jit(projection)
    # f_vg: Callable = jax.jit(jax.value_and_grad(f, argnums=0), **jit_options)
    # f_vg: Callable = jax.value_and_grad(f, argnums=0)
    f = jax.jit(f, **jit_options)
    grad: Callable = jax.jit(jax.grad(f, argnums=0), **jit_options)

    def _iter(tup: tuple, it):
        x: jnp.ndarray = tup[0]  # current estimate
        m: jnp.ndarray = tup[2]  # first moment estimate
        v: jnp.ndarray = tup[3]  # second moment estimate
        t: jnp.ndarray = tup[4]  # loop iteration count

        # getting raw gradient
        f_val: jnp.ndarray = f(x, **kwargs)
        f_grad: jnp.ndarray = grad(x, **kwargs)
        # f_val, f_grad = f_vg(x, **kwargs)
        f_grad = jnp.nan_to_num(f_grad)  # replace NaNs with 0s

        # performing Adam step
        d, m, v, t = adam(grad=f_grad, m=m, v=v, t=t, **adam_options)

        # performing projected gradient step
        x = single_update(x=x, d=d, lr=lr, projection=projection, projection_options=projection_options)
        # if jnp.isnan(f_val) or jnp.isnan(f_grad).any() or jnp.isnan(d).any() or jnp.isnan(x).any():
        #     raise ValueError("NaNs detected in the optimization loop.")
        return (x, f_val, m, v, t), it
    

    # initialise the optimization loop
    m0: jnp.ndarray = jnp.zeros_like(x0)
    v0: jnp.ndarray = jnp.zeros_like(x0)
    t: int = 0
    init = x0, jnp.inf, m0, v0, t

    # running projected gradient descent loop
    res, _ = jax.lax.scan(_iter, init, None, length=maxiter)
    # res = init
    # for i in range(maxiter):
    #     res, _ = _iter(res, i)

    # getting optimal values
    x_opt = res[0]
    # val_opt, _ = f_vg(x_opt, **kwargs)
    val_opt = f(x_opt, **kwargs)
    return {'x': x_opt, 'val': val_opt}


###############################################################################
# Brent's method
###############################################################################
@jax.jit
def _brent_new_bounds(a: Scalar, b: Scalar, c: Scalar, ga: Scalar, gb: Scalar, gc: Scalar) -> Array:
    # looking for same sign
    index = jnp.argmax(jnp.array([ga == 0, gb == 0, gc == 0, ga * gc > 0, ga * gc < 0], dtype=int))
    return jax.lax.switch(index, [
        lambda: jnp.array([a, a]),  # ga == 0
        lambda: jnp.array([b, b]),  # gb == 0
        lambda: jnp.array([c, c]),  # gc == 0
        lambda: jnp.array([c, b]),  # ga * gc > 0
        lambda: jnp.array([a, c])])


def _brentb(g: Callable, bounds: jnp.ndarray, **kwargs) -> Scalar:
    r"""Brent's bisection method for root finding.

    Args:
        g: function to find the root of.
        bounds: lower and upper bounds.
        maxiter: maximum number of iterations.

    Returns:
        dictionary containing optimal results.
    """
    a, b = bounds
    c: Scalar = 0.5 * (a + b)

    # evaluating the function at each point
    ga: Scalar = g(a, **kwargs)
    gb: Scalar = g(b, **kwargs)
    gc: Scalar = g(c, **kwargs)

    # returning new bounds
    return _brent_new_bounds(a=a, b=b, c=c, ga=ga, gb=gb, gc=gc)


@jax.jit
def _secant_root(a: Scalar, b: Scalar, ga: Scalar, gb: Scalar) -> Scalar:
    return b - gb * (b - a) / (gb - ga)


def _brents(g: Callable, bounds: jnp.ndarray, **kwargs) -> Scalar:
    r"""Brent's secant method for root finding.

    Args:
        g: function to find the root of.
        bounds: lower and upper bounds.
        maxiter: maximum number of iterations.

    Returns:
        dictionary containing optimal results.
    """
    a, b = bounds

    # evaluating the function at the bounds
    ga: Scalar = g(a, **kwargs)
    gb: Scalar = g(b, **kwargs)

    # deriving the root of the line
    c: Scalar = _secant_root(a=a, b=b, ga=ga, gb=gb)
    gc: Scalar = g(c, **kwargs)

    # returning new bounds
    return _brent_new_bounds(a=a, b=b, c=c, ga=ga, gb=gb, gc=gc)


def _brentq(g: Callable, bounds: jnp.ndarray, **kwargs) -> Scalar:
    r"""Brent's inverse quadratic method for root finding.

    Args:
        g: function to find the root of.
        bounds: lower and upper bounds.
        maxiter: maximum number of iterations.

    Returns:
        dictionary containing optimal results.
    """
    a, b = bounds

    # evaluating the function at the bounds
    ga: Scalar = g(a, **kwargs)
    gb: Scalar = g(b, **kwargs)

    # secant root
    c: Scalar = _secant_root(a=a, b=b, ga=ga, gb=gb)
    gc: Scalar = g(c, **kwargs)

    # finding root of the quadratic
    xq: Scalar = (
        (a * gb * gc) / ((ga - gb) * (ga - gc))
        + (b * ga * gc) / ((gb - ga) * (gb - gc))
        + (c * ga * gb) / ((gc - ga) * (gc - gb))
        )
    gxq: Scalar = g(xq, **kwargs)

    # returning new bounds
    # calc widths between all
    # find points with opposite signs
    # pick the two with smallest width
    combinations = [[a, c, ga, gc], [a, xq, ga, gxq], [b, c, gb, gc], 
                   [b, xq, gb, gxq], [c, xq, gc, gxq]]
    min_width = jnp.abs(b - a)
    for comb in combinations:
        i, j, gi, gj = comb
        width = jnp.abs(j - i)
        
        smaller_interval = jnp.logical_and(width < min_width, gi * gj < 0)
        index = jnp.argmax(jnp.array([gi == 0, gj ==0, smaller_interval, jnp.logical_not(smaller_interval)], dtype=int))
        min_width, bounds = jax.lax.switch(index, [
            lambda: (jnp.array(0.0), jnp.array([i, i])),  # gi == 0
            lambda: (jnp.array(0.0), jnp.array([j, j])),  # gj == 0
            lambda: (width, jnp.array([i, j])),  # smaller interval
            lambda: (min_width, bounds),  # default interval
        ])

    return jnp.sort(bounds)


def _brentsb(g: Callable, bounds: jnp.ndarray, **kwargs) -> Scalar:
    # bisection method
    bisection_bounds = _brentb(g=g, bounds=bounds, **kwargs)

    # secant method
    secant_bounds = _brents(g=g, bounds=bisection_bounds, **kwargs)
    return secant_bounds


def _brentqb(g: Callable, bounds: jnp.ndarray, **kwargs) -> Scalar:
    # bisection method
    bisection_bounds = _brentb(g=g, bounds=bounds, **kwargs)

    # quadratic method 
    quadratic_bounds = _brentq(g=g, bounds=bisection_bounds, **kwargs)
    return quadratic_bounds


def brent(g: Callable, bounds: jnp.ndarray, method: str = 'quadratic-bisection', 
          maxiter: int = 50, **kwargs) -> Scalar:
    r"""Brent's method for root finding.

    Args:
        g: function to find the root of.
        bounds: lower and upper bounds.
        method: method to use for root finding. Can be 'bisection',
            'secant', 'quadratic', 'secant-bisection', 'quadratic-bisection'.
        maxiter: maximum number of iterations.
        kwargs: additional arguments to pass to g.

    Returns:
        dictionary containing optimal results.
    """
    # getting method
    method = method.lower()
    if method == 'bisection':
        brent_method = _brentb
    elif method == 'secant':
        brent_method = _brents
    elif method == 'quadratic':
        brent_method = _brentq
    elif method == 'secant-bisection':
        brent_method = _brentsb
    elif method == 'quadratic-bisection':
        brent_method = _brentqb
    else:
        raise ValueError(f"Unknown method: {method}")

    # standardizing the bounds
    bounds: Array = jnp.asarray(bounds).flatten()
    bounds = jnp.sort(bounds)

    # iterating to find the root
    scan_func: Callable = lambda bounds_, _: (brent_method(g=g, bounds=bounds_, **kwargs), None)
    bounds, _ = jax.lax.scan(scan_func, bounds, None, length=maxiter)
    return bounds.mean()









"""Tests for copula distributions."""
import pytest
import jax.numpy as jnp
from jax import grad, jit
import numpy as np

from copulax._src._distributions import Univariate, Multivariate
from copulax._src.copulas._distributions import Copula
from copulax.tests.helpers import *
from copulax.tests.copulas.conftest import NUM_ASSETS
from copulax._src.typing import Scalar
from copulax.copulas import gaussian_copula, student_t_copula, skewed_t_copula, gh_copula


# Test combinations 
DISTRIBUTIONS = tuple((gaussian_copula, student_t_copula))
# DISTRIBUTIONS = tuple((gaussian_copula, student_t_copula, 
#                        skewed_t_copula, gh_copula))
ERROR_CASES = tuple(('too_large_dim_sample',))
DATASETS = tuple(('uncorrelated_sample', 'correlated_sample'))
COMBINATIONS = tuple((dist, dataset) for dist in DISTRIBUTIONS for dataset in DATASETS)


# def gradients(func, params: dict, s, data):
#     """Calculate the gradients of the output."""
#     new_func = lambda x: func(x, params=params).sum()
#     grad_output = grad(new_func)(data)
#     assert no_nans(grad_output), f"{s} gradient contains NaNs"
#     assert is_finite(grad_output), f"{s} gradient contains non-finite values"


def _check_marginals(dist, dim, fitted_marginals):
    # Checking properties
    assert isinstance(fitted_marginals, dict), f"{dist} fitted_marginals is not a dictionary."
    assert len(fitted_marginals) == 1, f"{dist} fitted_marginals does not have length 1."
    assert 'marginals' in fitted_marginals, f"{dist} fitted_marginals does not have 'marginals' key."
    marginals: tuple = fitted_marginals['marginals']
    assert isinstance(marginals, tuple), f"{dist} fitted_marginals subdict is not a tuple."
    fitted_num_dims = len(marginals)
    assert fitted_num_dims == dim, f"{dist} fitted_marginals has not fitted the correct number of dimensions. Expected {dim}, got {fitted_num_dims}."
    for marginal_tup in marginals:
        assert isinstance(marginal_tup, tuple), f"{dist} underlying fitted_marginals are not in tuple form."
        assert len(marginal_tup) == 2, f"{dist} underlying fitted_marginals tuples do not have length 2."
        marginal_dist, marginal_params = marginal_tup
        assert isinstance(marginal_dist, Univariate), f"{dist} underlying marginals distribution is not a Univariate object."
        assert isinstance(marginal_params, dict), f"{dist} underlying marginals parameters are not in dictionary form."
        check_uvt_params(marginal_params, f"{dist} underlying marginals parameters {marginal_params} are not valid.")


def _check_copula_params(dist, fitted_copula):
    # Checking properties
    assert isinstance(fitted_copula, dict), f"{dist} fitted_copula is not a dictionary."
    assert len(fitted_copula) == 1, f"{dist} fitted_copula does not have length 1."
    assert 'copula' in fitted_copula, f"{dist} fitted_copula does not have 'copula' key."
    copula_params = fitted_copula['copula']
    check_mvt_params(copula_params, f"{dist} copula_params")


# Tests for copula distributions
@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_all_objects(dist):
    assert isinstance(dist, Copula), f"{dist} is not a Copula object"
    assert isinstance(dist._mvt, Multivariate), f"{dist} mvt object is not a Multivariate object"
    assert isinstance(dist._uvt, Univariate), f"{dist} uvt object is not a Univariate object"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_all_methods_implemented(dist):
    methods: set[str] = {
        'dtype', 'dist_type', 'name', 'support', 'get_u', 'get_x_dash', 
        'copula_logpdf', 'copula_pdf', 'logpdf', 'pdf', 'copula_rvs', 
        'copula_sample', 'rvs', 'sample', 'fit_marginals', 'fit_copula', 
        'fit', 'aic', 'bic', 'loglikelihood', 'stats', 'example_params',
    }
    # testing desired methods are implemented
    for method in methods:
        assert hasattr(dist, method), f"{dist} missing method {method}"

    # testing no additional methods are implemented
    pytree_methods: set[str] = {'tree_flatten', 'tree_unflatten'}
    extra_methods = set(dist.__dict__.keys()) - methods - pytree_methods
    extra_methods = {m for m in extra_methods if not m.startswith('_')}
    assert not extra_methods, f"{dist} has extra methods: {extra_methods}"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_name(dist):
    assert isinstance(dist.name, str), f"{dist} name is not a string"
    assert dist.name != "", f"{dist} name is an empty string"
    assert dist.name == str(dist), f"{dist} name does not match its string representation"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dtype(dist):
    assert isinstance(dist.dtype, str), f"{dist} dtype is not a string"
    assert dist.dtype != "", f"{dist} dtype is an empty string"
    assert dist.dtype in ('continuous', 'discrete'), f"dtype is not 'continuous' or 'discrete' for {dist}"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dist_type(dist):
    assert isinstance(dist.dist_type, str), f"{dist} dist_type is not a string"
    assert dist.dist_type != "", f"{dist} dist_type is an empty string"
    assert dist.dist_type == 'copula', f"dist_type is not 'copula' for {dist}"


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_example_params(dist):
    params = dist.example_params(dim=NUM_ASSETS)

    # checking properties
    s = f"{dist} example_params"
    assert len(params) == 2 and 'marginals' in params and 'copula' in params, f"{s} does not have 'marginals' and 'copula' keys"
    
    # Check copula parameters
    copula_params = params.copy()
    copula_params.pop('marginals')
    _check_copula_params(dist, copula_params)

    # Check marginals parameters
    marginal_params = params.copy()
    marginal_params.pop('copula')
    _check_marginals(dist, NUM_ASSETS, marginal_params)


    for marginal, params in params['marginals']:
        assert isinstance(marginal, Univariate), f"{s} marginal {marginal} is not a Univariate object"
        check_uvt_params(params, f"{s} marginal {marginal} does not have valid params: {params}")


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_dist_object_jitable(dist):
    jittable(dist)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_support(dist):
    params = dist.example_params()
    support = dist.support(params)

    # Checking properties
    assert isinstance(support, jnp.ndarray), f"{dist} support is not a JAX array"
    assert two_dim(support), f"{dist} support is not a 2D array"
    assert np.all(support[:, 0] < support[:, 1]), f"{dist} support bounds are not in order."
    assert no_nans(support), f"{dist} support contains NaNs."

    # Check jit
    jitted_support = jit(dist.support)(params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_get_u(dist, dataset, datasets):
    params = dist.example_params()
    sample = datasets[dataset]

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            u = dist.get_u(sample, params)
    else:
        # Check non-error cases
        u = dist.get_u(sample, params)
        # Checking properties
        assert isinstance(u, jnp.ndarray), f"{dist} u is not a JAX array"
        assert two_dim(u), f"{dist} u is not a 2D array"
        assert np.all(u >= 0) and np.all(u <= 1), f"{dist} u is not in [0, 1]"
        assert no_nans(u), f"{dist} u contains NaNs."

        # Check jit
        jitted_u = jit(dist.get_u)(sample, params)

        # Check gradients
        gradients(dist.get_u, f"{dist} u", sample, params)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_get_x_dash(dist, u_sample):
    params = dist.example_params()
    x_dash = dist.get_x_dash(u_sample, params)
    
    # Checking properties
    assert isinstance(x_dash, jnp.ndarray), f"{dist} x_dash is not a JAX array"
    assert two_dim(x_dash), f"{dist} x_dash is not a 2D array"
    assert no_nans(x_dash), f"{dist} x_dash contains NaNs."

    # Check jit
    jitted_x_dash = jit(dist.get_x_dash)(u_sample, params)

    # Check gradients
    gradients(dist.get_x_dash, f"{dist} x_dash", u_sample, params)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_copula_logpdf(dist, u_sample):
    params = dist.example_params()
    logpdf = dist.copula_logpdf(u_sample, params)

    # Checking properties
    correct_mvt_shape(u_sample, logpdf, dist, 'copula_logpdf')
    assert no_nans(logpdf), f"{dist} copula_logpdf contains NaNs."

    # Check jit
    jitted_logpdf = jit(dist.copula_logpdf)(u_sample, params)

    # Check gradients
    gradients(dist.copula_logpdf, f"{dist} copula_logpdf", u_sample, params)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_copula_pdf(dist, u_sample):
    params = dist.example_params()
    pdf = dist.copula_pdf(u_sample, params)

    # Checking properties
    correct_mvt_shape(u_sample, pdf, dist, 'copula_pdf')
    assert is_positive(pdf), f"{dist} copula_pdf is not positive."
    assert no_nans(pdf), f"{dist} copula_pdf contains NaNs."
    assert is_finite(pdf), f"{dist} copula_pdf contains non-finite values."

    # Check jit
    jitted_pdf = jit(dist.copula_pdf)(u_sample, params)

    # Check gradients
    gradients(dist.copula_pdf, f"{dist} copula_pdf", u_sample, params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_logpdf(dist, dataset, datasets):
    params = dist.example_params()
    sample = datasets[dataset]

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            logpdf = dist.logpdf(sample, params)

    else:
        logpdf = dist.logpdf(sample, params)

        # Checking properties
        correct_mvt_shape(sample, logpdf, dist, 'logpdf')
        assert no_nans(logpdf), f"{dist} logpdf contains NaNs."

        # Check jit
        jitted_logpdf = jit(dist.logpdf)(sample, params)

        # Check gradients
        gradients(dist.logpdf, f"{dist} logpdf", sample, params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_pdf(dist, dataset, datasets):
    params = dist.example_params()
    sample = datasets[dataset]

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            pdf = dist.pdf(sample, params)
    else:
        # Check non-error cases
        pdf = dist.pdf(sample, params)

        # Checking properties
        correct_mvt_shape(sample, pdf, dist, 'pdf')
        assert is_positive(pdf), f"{dist} pdf contains negative values."
        assert no_nans(pdf), f"{dist} pdf contains NaNs."
        assert is_finite(pdf), f"{dist} pdf contains non-finite values."

        # Check jit
        jitted_pdf = jit(dist.pdf)(sample, params)

        # Check gradients
        gradients(dist.pdf, f"{dist} pdf", sample, params)


SIZES = (0, 1, 2, 11)
SIZE_COMBINATIONS = tuple((dist, size) for dist in DISTRIBUTIONS for size in SIZES)
@pytest.mark.parametrize("dist, size", SIZE_COMBINATIONS)
def test_copula_rvs(dist, size):
    params = dist.example_params()
    rvs = dist.copula_rvs(size=size, params=params)

    # Checking properties
    assert two_dim(rvs), f"{dist} copula_rvs is not a 2D array"
    expected_shape = (size, NUM_ASSETS)
    assert rvs.shape == expected_shape, f"{dist} copula_rvs has incorrect shape. Expected {expected_shape}, got {rvs.shape}"
    assert no_nans(rvs), f"{dist} copula_rvs contains NaNs for size {size}"
    assert is_finite(rvs), f"{dist} copula_rvs contains infinite values for size {size}"
    
    # Check jit
    jitted_rvs = jit(dist.copula_rvs, static_argnums=0)(size=size, params=params)


@pytest.mark.parametrize("dist, size", SIZE_COMBINATIONS)
def test_rvs(dist, size):
    params = dist.example_params()
    rvs = dist.rvs(size=size, params=params)

    # Checking properties
    assert two_dim(rvs), f"{dist} rvs is not a 2D array"
    expected_shape = (size, NUM_ASSETS)
    assert rvs.shape == expected_shape, f"{dist} rvs has incorrect shape. Expected {expected_shape}, got {rvs.shape}"
    assert no_nans(rvs), f"{dist} rvs contains NaNs for size {size}"
    assert is_finite(rvs), f"{dist} rvs contains infinite values for size {size}"

    # Check jit
    jitted_rvs = jit(dist.rvs, static_argnums=0)(size=size, params=params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_fit_marginals(dist, dataset, datasets):
    sample = datasets[dataset]
    fitted_marginals = dist.fit_marginals(sample)

    # Checking properties
    _check_marginals(dist, sample.shape[1], fitted_marginals)


@pytest.mark.parametrize("dist", DISTRIBUTIONS)
def test_fit_copula(dist, u_sample):
    fitted_copula = dist.fit_copula(u_sample)
    # Checking properties
    _check_copula_params(dist, fitted_copula)
    # Check jit
    jitted_fitted_copula = jit(dist.fit_copula)(u_sample)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_fit(dist, dataset, datasets):
    sample = datasets[dataset]
    fitted_joint = dist.fit(sample)
    
    # Checking properties
    assert isinstance(fitted_joint, dict), f"{dist} fit does not return a dictionary"
    assert len(fitted_joint) == 2, f"{dist} fit does not have length 2."
    assert 'marginals' in fitted_joint and 'copula' in fitted_joint, f"{dist} fit does not have 'marginals' and 'copula' keys."
    fitted_marginals = fitted_joint.copy()
    fitted_marginals.pop('copula')
    _check_marginals(dist, sample.shape[1], fitted_marginals)
    fitted_copula = fitted_joint.copy()
    fitted_copula.pop('marginals')
    _check_copula_params(dist, fitted_copula)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_aic(dist, dataset, datasets):
    params = dist.example_params()
    sample = datasets[dataset]

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.aic(sample, params)
    else:
        output = dist.aic(sample, params)

        # Checking properties
        check_metric_output(dist, output, 'aic')

        # Check jit
        jitted_aic = jit(dist.aic)(sample, params)

        # Check gradients
        gradients(dist.aic, f"{dist} aic", sample, params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_bic(dist, dataset, datasets):
    params = dist.example_params()
    sample = datasets[dataset]

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.bic(sample, params)
    else:
        output = dist.bic(sample, params)

        # Checking properties
        check_metric_output(dist, output, 'bic')

        # Check jit
        jitted_bic = jit(dist.bic)(sample, params)

        # Check gradients
        gradients(dist.bic, f"{dist} bic", sample, params)


@pytest.mark.parametrize("dist, dataset", COMBINATIONS)
def test_loglikelihood(dist, dataset, datasets):
    params = dist.example_params()
    sample = datasets[dataset]

    if dataset in ERROR_CASES:
        # Check error cases - dim mismatch between data and params
        with pytest.raises(TypeError):
            output = dist.loglikelihood(sample, params)
    else:
        output = dist.loglikelihood(sample, params)

        # Checking properties
        check_metric_output(dist, output, 'loglikelihood')

        # Check jit
        jitted_loglikelihood = jit(dist.loglikelihood)(sample, params)

        # Check gradients
        gradients(dist.loglikelihood, f"{dist} loglikelihood", sample, params)



"""CopulAX implementation of the popular Copula distributions."""
from jax._src.typing import ArrayLike, Array
from typing import Callable
from collections import deque
from jax import numpy as jnp
from jax import jit

from copulax._src._distributions import GeneralMultivariate, Multivariate, Univariate, dist_map
from copulax.univariate import univariate_fitter
from copulax._src.multivariate._utils import _multivariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.typing import Scalar

from copulax._src.multivariate.mvt_normal import mvt_normal
from copulax._src.univariate.normal import normal
from copulax._src.multivariate.mvt_student_t import mvt_student_t
from copulax._src.univariate.student_t import student_t
from copulax._src.multivariate.mvt_gh import mvt_gh
from copulax._src.univariate.gh import gh
from copulax._src.multivariate.mvt_skewed_t import mvt_skewed_t
from copulax._src.univariate.skewed_t import skewed_t


###############################################################################
# Copula Class 
# (Must be located here to prevent circular import issues)
###############################################################################
class Copula(GeneralMultivariate):
    r"""Base class for copula distributions."""
    # initialisation
    def __init__(self, name, mvt: Multivariate, uvt: Univariate):
        super().__init__(name)
        self._mvt: Multivariate = mvt  # multivariate pytree object
        self._uvt: Univariate = uvt  # univariate pytree object

    @classmethod
    def tree_unflatten(cls, aux_data: tuple, values: tuple, **init_kwargs):
        id_: int = aux_data[0]
        return cls(dist_map.id_map[id_]['name'], *values, **init_kwargs)
    
    def tree_flatten(self):
        children = (self._mvt, self._uvt)  # arrays and pytrees
        aux_data = (self._id,)  # static, hashable data
        return children, aux_data
    
    # standard functions
    def _get_dim(self, params) -> int:
        return len(params['marginals'])

    def support(self, params: dict) -> Array:
        marginals: tuple = params['marginals']
        
        support: deque = deque()
        for dist, mparams in marginals:
            msupport = dist.support(params=mparams)
            support.append(msupport)
        
        return jnp.vstack(support)
    
    def _example_copula_params(self, dim, *args, **kwargs) -> dict:
        # override if sigma is not the shape matrix for the mulviariate dist
        mvt_params: dict = self._mvt.example_params(dim=dim, *args, **kwargs)
        mvt_params['sigma'] = jnp.eye(dim, dim)
        return mvt_params

    def example_params(self, dim: int = 3, *args, **kwargs):
        r"""Example parameters for the copula distribution.
        
        Generates example marginal and copula parameters for the overall 
        joint distribution.

        Args:
            dim: int, number of dimensions of the copula distribution. 
                Default is 3.
        """
        # copula parameters
        mvt_params: dict = self._mvt.example_params(dim=dim, *args, **kwargs)
        mvt_params['sigma'] = jnp.eye(dim, dim)

        # marginal parameters
        marginal_params: tuple = tuple((self._uvt, self._uvt.example_params(dim=dim)) for _ in range(dim)) 

        # joint parameters
        return {'marginals': marginal_params, 'copula': mvt_params}

    def get_u(self, x: ArrayLike, params: dict) -> Array:
        r"""Computes the independent univariate marginal cdf values (u) 
        for each dimension.
        
        Args:
            x (ArrayLike): The multivariate input data to evaluate the 
                cdf.
            params (dict): The marginal distribution parameters. 

        Returns:
            u (Array): The independent univariate marginal cdf values 
                for each dimension.
        """
        x: jnp.ndarray = _multivariate_input(x)[0]
        marginals: tuple = params['marginals']
        u: deque = deque()
        for i, (dist, mparams) in enumerate(marginals):
            ui: jnp.ndarray = dist.cdf(x[:, i][:, None], params=mparams)
            u.append(ui)
        return jnp.concat(u, axis=1)

    def _get_uvt_params(self, params: dict) -> tuple:
        """Returns the univariate distribution parameters."""
        return tuple()
        
    def _scan_uvt_func(self, func: Callable, x: Array, params: dict, **kwargs) -> Array:
        """Scans the univariate distribution function over the data."""
        uvt_params_tuple: tuple = self._get_uvt_params(params)
        d: int = self._get_dim(params)
        output: deque = deque(maxlen=d)
        for i, uvt_params in enumerate(uvt_params_tuple):
            output_i: Array = func(x[:, i][:, None], params=uvt_params, **kwargs)
            output.append(output_i)
        return jnp.concat(output, axis=1)


    def get_x_dash(self, u: ArrayLike, params: dict) -> Array:
        r"""Computes x' values, which represent the mappings of the 
        independent marginal cdf values (U) to the domain of the joint 
        multivariate distribution.
        
        Args:
            u (ArrayLike): The independent univariate marginal cdf 
                values (U) for each dimension.
            params (dict): The copula and marginal distribution 
                parameters.

        Returns:
            x_dash (Array): The x' values for each dimension.
        """
        u_raw: jnp.ndarray = _multivariate_input(u)[0]
        eps: float = 1e-4
        u: jnp.ndarray = jnp.clip(u_raw, eps, 1 - eps)
        # uvt_params: dict = self._get_uvt_params(params)
        # uvt_ppf: Callable = jit(self._uvt.ppf, static_argnames=('cubic',))
        # return uvt_ppf(u, params=uvt_params, cubic=True)
        uvt_ppf: Callable = jit(self._uvt.ppf, static_argnames=('cubic',))
        return self._scan_uvt_func(func=uvt_ppf, x=u, params=params, cubic=True)
    
    # densities
    def copula_logpdf(self, u: ArrayLike, params: dict) -> Array:
        r"""Computes the log-pdf of the copula distribution.
        
        Args:
            u (ArrayLike): The independent univariate marginal cdf 
                values (u) for each dimension.
            params (dict): The copula and marginal distribution 
                parameters.

        Returns:
            logpdf (Array): The log-pdf values of the copula 
                distribution.
        """
        # mapping u to x' space
        x_dash: jnp.ndarray = self.get_x_dash(u, params)

        # computing univariate logpdfs
        # uvt_params: dict = self._get_uvt_params(params)
        # uvt_logpdf: jnp.ndarray = self._uvt.logpdf(x_dash, params=uvt_params)
        uvt_logpdf: jnp.ndarray = self._scan_uvt_func(
            func=jit(self._uvt.logpdf), x=x_dash, params=params)
        
        # computing copula logpdf
        mvt_params: dict = params['copula']
        mvt_logpdf: jnp.ndarray = self._mvt.logpdf(x_dash, params=mvt_params)
        return mvt_logpdf - uvt_logpdf.sum(axis=1, keepdims=True)
    
    def copula_pdf(self, u: ArrayLike, params: dict) -> Array:
        r"""Computes the pdf of the copula distribution.

        Args:
            u (ArrayLike): The independent univariate marginal cdf 
                values (U) for each dimension.
            params (dict): The copula and marginal distribution 
                parameters.

        Returns:
            pdf (Array): The pdf values of the copula distribution.
        """
        return jnp.exp(self.copula_logpdf(u, params))
    
    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        r"""The log-probability density function (pdf) of the overall 
        joint copula and marginal distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the log-pdf.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.
            params (dict): The copula and marginal distribution 
                parameters.

        Returns:
            Array: The log-pdf values of the overall joint copula and 
                marginal distribution.
        """
        # calculating marginal logpdfs
        x, _, n, d = _multivariate_input(x)
        marginals: tuple = params['marginals']
        marginal_logpdf_sum: jnp.ndarray = jnp.zeros((n, 1))
        for i, (dist, mparams) in enumerate(marginals):
            marginal_logpdf_sum += dist.logpdf(x[:, i][:, None], params=mparams)

        # calculating copula logpdf
        u: jnp.ndarray = self.get_u(x, params)
        copula_logpdf: jnp.ndarray = self.copula_logpdf(u, params)
        return copula_logpdf + marginal_logpdf_sum

    def pdf(self, x: ArrayLike, params: dict) -> Array:
        r"""The probability density function (pdf) of the overall joint 
        copula and marginal distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the pdf.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.
            params (dict): The copula and marginal distribution 
                parameters.

        Returns:
            Array: The pdf values of the overall joint copula
                and marginal distribution.
        """
        return jnp.exp(self.logpdf(x, params))

    # sampling
    def copula_rvs(self, size: Scalar, params: dict, 
                   key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the copula distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (Scalar): size (Scalar): The size / shape of the generated 
                output array of random numbers. Must be scalar. 
                Generates an (size, d) array of random numbers, where
                d is the number of dimensions inferred from the provided
                distribution parameters.
            params (dict): The copula and marginal distribution 
                parameters.
            key (Array): The Key for random number generation.
        """
        # generating random samples from x'
        x_dash: jnp.ndarray = self._mvt.rvs(size=size, key=key, params=params['copula'])

        # projecting x' to u space
        return self._scan_uvt_func(jit(self._uvt.cdf), x=x_dash, params=params)
        # uvt_params: dict = self._get_uvt_params(params)
        # return self._uvt.cdf(x_dash, params=uvt_params)

    def copula_sample(self, size: Scalar, params: dict, 
                   key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the copula distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (Scalar): size (Scalar): The size / shape of the generated 
                output array of random numbers. Must be scalar. 
                Generates an (size, d) array of random numbers, where
                d is the number of dimensions inferred from the provided
                distribution parameters.
            params (dict): The copula and marginal distribution 
                parameters.
            key (Array): The Key for random number generation.
        """
        return self.copula_rvs(size=size, key=key, params=params)
    
    def rvs(self, size: Scalar, params: dict, 
            key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the overall joint copula and 
        marginal distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (Scalar): size (Scalar): The size / shape of the generated 
                output array of random numbers. Must be scalar. 
                Generates an (size, d) array of random numbers, where
                d is the number of dimensions inferred from the provided
                distribution parameters.
            params (dict): The copula and marginal distribution 
                parameters.
            key (Array): The Key for random number generation.
        """
        # sampling from copula distribution
        u_raw: jnp.ndarray = self.copula_rvs(size=size, key=key, params=params)
        eps: float = 1e-4
        u: jnp.ndarray = jnp.clip(u_raw, eps, 1 - eps)
        
        # projecting u to x space
        marginals: tuple = params['marginals']
        x: deque = deque()
        
        for i, (dist, mparams) in enumerate(marginals):
            xi: jnp.ndarray = dist.ppf(u[:, i][:, None], params=mparams, 
                                       cubic=True)
            x.append(xi)
        
        return jnp.concat(x, axis=1)

    def sample(self, size: Scalar, params: dict, 
            key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the overall joint copula and 
        marginal distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (Scalar): size (Scalar): The size / shape of the generated 
                output array of random numbers. Must be scalar. 
                Generates an (size, d) array of random numbers, where
                d is the number of dimensions inferred from the provided
                distribution parameters.
            params (dict): The copula and marginal distribution 
                parameters.
            key (Array): The Key for random number generation.
        """
        return self.rvs(size=size, key=key, params=params)

    # fitting
    def fit_marginals(self, x: ArrayLike, 
                      univariate_fitter_options: tuple[dict] | dict = None, 
                      ) -> dict:
        r"""Fits the univariate marginal distributions to the data.

        Args:
            x (ArrayLike): The input data to fit the distribution too.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.
            univariate_fitter_options (tuple | dict): The options for 
                the univariate fitter. If a dictionary is provided, 
                specify the desired univariate argument names and values 
                as key value pairs. These will then be used for each 
                distribution. If a tuple is provided, it must contain 
                dictionaries for each distribution using the same 
                indexing as in 'x'. 

        Note:
            Not jitable.

        Returns:
            dict: A params dict containing the fitted univariate 
            marginal distributions and their parameters.
        """
        # determining univariate fitter options
        x, _, n, d = _multivariate_input(x)
        if univariate_fitter_options is None:
            univariate_fitter_options: tuple = ({},) * d
        elif isinstance(univariate_fitter_options, dict):
            univariate_fitter_options: tuple = (univariate_fitter_options,) * d
        elif isinstance(univariate_fitter_options, tuple):
            if len(univariate_fitter_options) != d:
                raise ValueError(f"If univariate_fitter_options is a tuple, "
                                 "it must have an entry for each variable in x.")
        else:
            raise ValueError(f"univariate_fitter_options must be a tuple or "
                             "dictionary.")
        
        # fitting marginals
        jitted_ufitter: Callable = jit(univariate_fitter, static_argnames=('metric', 'distributions'))
        marginals: deque = deque(maxlen=d)
        for i, options in enumerate(univariate_fitter_options):
            xi: jnp.ndarray = x[:, i]
            best_index, fitted = jitted_ufitter(xi, **options)
            dist: Univariate = fitted[best_index]['dist']
            params: dict = fitted[best_index]['params']
            marginals.append((dist, params))
        
        return {'marginals':  tuple(marginals)}
    
# """            marginals (dict): The marginal distribution's and their 
#                 parameters. Must be a dictionary with a 'marginals' key 
#                 with a tuple value; this tuple must contain 'd' tuples 
#                 (one for each variable), with a univariate distribution 
#                 and their dictionary parameters. 
#                 Obtainable by calling 'fit_marginals'."""

    def fit_copula(self, u: ArrayLike, corr_method: str = 'pearson', 
                   lr: float = 1e-4, maxiter: int = 100) -> dict:
        r"""Fits the copula parameters to the data.

        Args:
            u (ArrayLike): The independent univariate marginal cdf 
                values (u) for each dimension.
            corr_method: str, method to estimate the sample correlation 
                matrix, sigma. See copulax.multivariate.corr 
                for available methods. Used for copula's which require 
                a correlation matrix to be estimated, namely Gaussian,
                Student-T, Skewed-T and GH copulas.
            lr (float): Learning rate for optimization. Used for 
                copula's which require numerical methods for parameter
                estimation, namely Student-T, Skewed-T and GH copulas.
            maxiter (int): Maximum number of iterations for optimization.
                Used for copula's which require numerical methods for 
                parameter estimation, namely Student-T, Skewed-T and GH 
                copulas.

        Returns:
            dict: A params dict containing the fitted copula parameters.
        """
        # fitting copula
        copula: dict = self._mvt._fit_copula(u, corr_method=corr_method, 
                                             lr=lr, maxiter=maxiter)
        return {'copula': copula}

    def fit(self, x: ArrayLike, univariate_fitter_options: tuple[dict] | dict = None, 
            corr_method: str = 'pearson', lr: float = 1e-4, maxiter: int = 100) -> dict:
        r"""Fits the joint copula and marginal distribution to the data.
        This is equivalent to calling 'fit_marginals' and 'fit_copula' 
        successively.

        Args:
            x (ArrayLike): The input data to fit the distribution too.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.
            univariate_fitter_options (tuple | dict): The options for
                the univariate fitter. If a dictionary is provided, 
                specify the desired univariate argument names and values 
                as key value pairs. These will then be used for each 
                distribution. If a tuple is provided, it must contain 
                dictionaries for each distribution using the same 
                indexing as in 'x'.
            corr_method: str, method to estimate the sample correlation 
                matrix, sigma. See copulax.multivariate.corr 
                for available methods. Used for copula's which require 
                a correlation matrix to be estimated, namely Gaussian,
                Student-T, Skewed-T and GH copulas.
            lr (float): Learning rate for optimization. Used for 
                copula's which require numerical methods for parameter
                estimation, namely Student-T, Skewed-T and GH copulas.
            maxiter (int): Maximum number of iterations for optimization.
                Used for copula's which require numerical methods for 
                parameter estimation, namely Student-T, Skewed-T and GH 
                copulas.

        Note:
            Not jitable.

        Returns:
            dict: A params dict containing the fitted univariate 
                marginal distributions and their parameters, in addition 
                to the fitted copula parameters.
        """
        # fitting marginals
        marginals: dict = self.fit_marginals(x, univariate_fitter_options)

        # projecting x to u space
        u: jnp.ndarray = self.get_u(x, marginals)

        # fitting copula
        copula: dict = self.fit_copula(u, corr_method=corr_method, 
                                        lr=lr, maxiter=maxiter)
        
        # fitting joint distribution
        return {**marginals, **copula}

    # metrics
    # get num params -> rest (loglikelihood, aic, bic) ss follow from inheritience

###############################################################################
# Copula Distributions
###############################################################################
# Normal Mixture Copulas
class GaussianCopula(Copula):
    r"""The Gaussian Copula is a copula that uses the multivariate normal
    distribution to model the dependencies between random variables.
    
    https://en.wikipedia.org/wiki/Copula_(statistics)
    """
    @jit
    def _get_uvt_params(self, params: dict) -> dict:
        d: int = self._get_dim(params)
        return tuple(({"mu": 0.0, "sigma": 1.0},) * d)


gaussian_copula = GaussianCopula('Gaussian-Copula', mvt_normal, normal)


class StudentTCopula(Copula):
    r"""The Student-T Copula is a copula that uses the multivariate 
    Student-T distribution to model the dependencies between random 
    variables.
    
    https://en.wikipedia.org/wiki/Copula_(statistics)
    """
    @jit
    def _get_uvt_params(self, params: dict) -> dict:
        nu: Scalar = params["copula"]["nu"]
        d: int = self._get_dim(params)
        return tuple(({"nu": nu, "mu": 0.0, "sigma": 1.0},) * d)

student_t_copula = StudentTCopula('Student-T-Copula', mvt_student_t, student_t)


class GHCopula(Copula):
    r"""The GH Copula is a copula that uses the multivariate generalized 
    hyperbolic (GH) distribution to model the dependencies between 
    random variables.

    https://en.wikipedia.org/wiki/Copula_(statistics)
    """
    @jit
    def _get_uvt_params(self, params: dict) -> dict:
        lamb: Scalar = params["copula"]["lambda"]
        chi: Scalar = params["copula"]["chi"]
        psi: Scalar = params["copula"]["psi"]
        gamma: Array = params["copula"]["gamma"]

        uvt_params: deque = deque()
        for i, gamma_i in enumerate(gamma.flatten()):
            params_i: dict = {"lambda": lamb, "chi": chi, "psi": psi, 
                              "mu": 0.0, "sigma": 1.0, "gamma": gamma_i}
            uvt_params.append(params_i)
        return tuple(uvt_params)
    

gh_copula = GHCopula('GH-Copula', mvt_gh, gh)


class SkewedTCopula(Copula):
    r"""The Skewed-T Copula is a copula that uses the multivariate 
    skewed-T distribution to model the dependencies between random 
    variables.

    https://en.wikipedia.org/wiki/Copula_(statistics)
    """
    def _get_uvt_params(self, params: dict) -> dict:
        nu: Scalar = params["copula"]["nu"]
        gamma: Array = params["copula"]["gamma"]

        uvt_params: deque = deque()
        for i, gamma_i in enumerate(gamma.flatten()):
            params_i: dict = {"nu": nu, "mu": 0.0, 
                              "sigma": 1.0, "gamma": gamma_i}
            uvt_params.append(params_i)
        return tuple(uvt_params)


skewed_t_copula = SkewedTCopula('Skewed-T-Copula', mvt_skewed_t, skewed_t)


"""File containing the copulAX implementation of the student-T distribution."""
import jax.numpy as jnp
from jax import lax, custom_vjp, random
from jax.scipy import special
from jax._src.typing import ArrayLike, Array
from copy import deepcopy

from copulax._src._distributions import Univariate
from copulax._src.typing import Scalar
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.univariate._cdf import _cdf, cdf_bwd, _cdf_fwd
from copulax._src.optimize import projected_gradient


class StudentTBase(Univariate):
    @classmethod
    def _params_dict(cls, nu: Scalar, mu: Scalar, sigma: Scalar) -> dict:
        d: dict = {"nu": nu, "mu": mu, "sigma": sigma}
        return cls._args_transform(d)
    
    @staticmethod
    def _params_to_tuple(params: dict) -> tuple:
        params = StudentTBase._args_transform(params)
        return params["nu"], params["mu"], params["sigma"]
    
    @staticmethod
    def _params_to_array(params: dict) -> Array:
        return jnp.asarray(StudentTBase._params_to_tuple(params)).flatten()
    
    def example_params(self, *args, **kwargs) -> dict:
        r"""Example parameters for the student-T distribution.
        
        This is a three parameter family, with the student-T being defined by 
        its degrees of freedom `nu`, location `mu` and scale `sigma`.
        """
        return self._params_dict(nu=2.5, mu=0.0, sigma=1.0)
    
    @staticmethod
    def _support(*args, **kwargs) -> Array:
        return -jnp.inf, jnp.inf
    
    @staticmethod
    def _stable_logpdf(stability: Scalar, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        nu, mu, sigma = StudentTBase._params_to_tuple(params)

        z: jnp.ndarray = lax.div(lax.sub(x, mu), sigma)

        const: jnp.ndarray = special.gammaln(0.5 * (nu + 1)) - special.gammaln(0.5 * nu) - 0.5 * lax.log(stability + (nu * jnp.pi * sigma))
        e: jnp.ndarray = lax.mul(lax.log(stability + lax.add(1.0, lax.div(lax.pow(z, 2.0), nu))), -0.5 * (nu + 1))
        logpdf = lax.add(const, e)
        return logpdf.reshape(xshape)
    
    @staticmethod
    def logpdf(x: ArrayLike, params: dict) -> Array:
        return StudentTBase._stable_logpdf(stability=0.0, x=x, params=params)
    
    @staticmethod
    def pdf(x: ArrayLike, params: dict) -> Array:
        return jnp.exp(StudentTBase.logpdf(x=x, params=params))
    
    # ppf
    # def _get_x0(self, params: dict) -> Scalar:
    #     return self._args_transform(params)["mu"]

    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, 
            key: Array = DEFAULT_RANDOM_KEY) -> Array:
        nu, mu, sigma = self._params_to_tuple(params)
        z: jnp.ndarray = random.t(key=key, df=nu, shape=size)
        return lax.add(lax.mul(z, sigma), mu)
    
    # stats
    def stats(self, params: dict) -> dict:
        nu, mu, sigma = self._params_to_tuple(params)
        mean: float = jnp.where(nu > 1, mu, jnp.nan)
        variance: float = jnp.where(nu > 2, nu / (nu - 2), jnp.nan)
        std: float = jnp.sqrt(variance)
        skewness: float = jnp.where(nu > 3, 0.0, jnp.nan)
        kurtosis: float = jnp.where(nu > 4, 6 / (nu - 4), jnp.inf)
        kurtosis = jnp.where(nu <= 2, jnp.nan, kurtosis)

        return self._scalar_transform({
            "mean": mean, 
            "median": mu, 
            "mode": mu, 
            "variance": variance, 
            "std": std, 
            "skewness": skewness, 
            "kurtosis": kurtosis})
    
    # fitting
    def _fit_mle(self, x: jnp.ndarray, lr: float, maxiter: int) -> dict:
        eps = 1e-8
        constraints: tuple = (jnp.array([[eps, -jnp.inf, eps]]).T, 
                            jnp.array([[jnp.inf, jnp.inf, jnp.inf]]).T)
        
        projection_options: dict = {'hyperparams': constraints}
        params0: jnp.ndarray = jnp.array([
            jnp.exp(random.normal(key=DEFAULT_RANDOM_KEY) * 2.5), 
            x.mean(), x.std()])

        res = projected_gradient(
            f=self._mle_objective, x0=params0, projection_method='projection_box', 
            projection_options=projection_options, x=x, lr=lr, maxiter=maxiter)
        nu, mu, sigma = res['x']

        return self._params_dict(nu=nu, mu=mu, sigma=sigma)
    
    def _ldmle_objective(self, params_arr: jnp.ndarray, x: jnp.ndarray, sample_mean: Scalar) -> jnp.ndarray:
        nu, sigma = params_arr
        return self._mle_objective(params_arr=jnp.array([nu, sample_mean, sigma]), x=x)
    
    def _fit_ldmle(self, x: jnp.ndarray, lr: float, maxiter: int) -> dict:
        params0: jnp.ndarray = jnp.array([1.0, x.std()])
        sample_mean: float = x.mean()
        res = projected_gradient(
            f=self._ldmle_objective, x0=params0, 
            projection_method='projection_non_negative', x=x, 
            sample_mean=sample_mean, lr=lr, maxiter=maxiter)
        nu, sigma = res['x']

        return self._params_dict(nu=nu, mu=sample_mean, sigma=sigma)
    
    def fit(self, x: ArrayLike, method: str = 'LDMLE', lr: float = 0.1, 
            maxiter: int = 100,) -> dict:
        r"""Fit the distribution to the input data.

        Note:
            If you intend to jit wrap this function, ensure that 'method' is a 
            static argument.
        
        Args:
            x (ArrayLike): The input data to fit the distribution to.
            method (str): The fitting method to use.  Options are 
                'MLE' for maximum likelihood estimation, and 'LDMLE' 
                for low-dimensional maximum likelihood estimation. 
                Defaults to 'LDMLE'. 
            lr (float): Learning rate for the fitting process.
            maxiter (int): Maximum number of iterations for the fitting process.
        
        Returns:
            dict: The fitted distribution parameters.
        """
        x = _univariate_input(x)[0]
        if method == 'MLE':
            return self._fit_mle(x, lr=lr, maxiter=maxiter)
        else:
            return self._fit_ldmle(x, lr=lr, maxiter=maxiter)
        
    # cdf
    # @staticmethod
    # def _params_from_array(params_arr, *args, **kwargs):
    #     nu, mu, sigma = params_arr
    #     return StudentTBase._params_dict(nu=nu, mu=mu, sigma=sigma)

    # @staticmethod
    # def _pdf_for_cdf(x: ArrayLike, *params_tuple) -> Array:
    #     params_array: jnp.ndarray = jnp.asarray(params_tuple).flatten()
    #     params: dict = StudentTBase._params_from_array(params_array)
    #     return StudentTBase.pdf(x=x, params=params)
    
def _vjp_cdf(x: ArrayLike, params: dict) -> Array:
    params = StudentTBase._args_transform(params)
    return _cdf(dist=StudentTBase, x=x, params=params)


_vjp_cdf_copy = deepcopy(_vjp_cdf)
_vjp_cdf = custom_vjp(_vjp_cdf)


def cdf_fwd(x: ArrayLike, params: dict) -> tuple[Array, tuple]:
    params = StudentTBase._args_transform(params)
    return _cdf_fwd(dist=StudentTBase, cdf_func=_vjp_cdf_copy, x=x, params=params)


_vjp_cdf.defvjp(cdf_fwd, cdf_bwd)


class StudentT(StudentTBase):
    r"""The student-T distribution is a 3 parameter family of continuous 
    distributions which generalize the normal distribuion, allowing it to have 
    heavier tails.

    https://en.wikipedia.org/wiki/Student%27s_t-distribution
    """
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        return _vjp_cdf(x=x, params=params)
    

student_t = StudentT("Student-T")


"""File containing the copulAX implementation of the Inverse Gaussian distribution."""
import jax.numpy as jnp
from jax import lax, random, scipy
from jax._src.typing import ArrayLike, Array
from tensorflow_probability.substrates import jax as tfp

from copulax._src._distributions import Univariate
from copulax._src.typing import Scalar
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.optimize import projected_gradient
from copulax._src.univariate.gamma import gamma


class IG(Univariate):
    r"""The inverse gamma distribution is a two-parameter family of continuous 
    probability distributions which represents the reciprocal of gamma distributed 
    random variables.

    We use the rate parameterization of the inverse gamma distribution specified by
    McNeil et al (2005).

    https://en.wikipedia.org/wiki/Inverse-gamma_distribution
    """
    @classmethod
    def _params_dict(cls, alpha: Scalar, beta: Scalar) -> dict:
        d: dict = {"alpha": alpha, "beta": beta}
        return cls._args_transform(d)

    def _params_to_tuple(self, params):
        params = self._args_transform(params)
        return params["alpha"], params["beta"]
    
    def example_params(self, *args, **kwargs) -> dict:
        r"""Example parameters for the inverse gamma distribution.
        
        This is a two parameter family, with the inverse gamma being 
        defined by parameters `alpha` and `beta`.
        """
        return self._params_dict(alpha=1.0, beta=1.0)

    @classmethod
    def _support(cls, *args, **kwargs) -> tuple:
        return 0.0, jnp.inf
    
    def _stable_logpdf(self, stability: Scalar, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        alpha, beta = self._params_to_tuple(params)
        
        logpdf: jnp.ndarray = (alpha * jnp.log(beta + stability) 
                               - lax.lgamma(alpha) 
                               - (alpha + 1) * jnp.log(x) 
                               - (beta / x))
        return logpdf.reshape(xshape)

    def cdf(self, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        alpha, beta = self._params_to_tuple(params)
        cdf: jnp.ndarray = scipy.special.gammaincc(a=alpha, x=(beta / x))
        return cdf.reshape(xshape)
    
    # ppf
    # def _get_x0(self, params):
    #     return self.stats(params=params)['mode']

    def _ppf(self, q: ArrayLike, params: dict, *args, **kwargs) -> Array:
        alpha, beta = self._params_to_tuple(params)
        return beta / tfp.math.igammacinv(a=alpha, p=q)
    
    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, key: Array = DEFAULT_RANDOM_KEY) -> Array:
        return 1.0 / gamma.rvs(size=size, key=key, params=params)
    
    # stats
    def stats(self, params: dict) -> dict:
        alpha, beta = self._params_to_tuple(params)
        mean: float = jnp.where(alpha > 1.0, beta / (alpha - 1), jnp.nan)
        mode: float = beta / (alpha + 1)
        variance: float = jnp.where(alpha > 2.0, lax.pow(beta, 2) / (lax.pow(alpha - 1, 2) * (alpha - 1)), jnp.nan)
        std: float = jnp.sqrt(variance)
        skewness: float = jnp.where(alpha > 3.0, 4 * jnp.sqrt(alpha - 2) / (alpha - 3), jnp.nan)
        kurtosis: float = jnp.where(alpha > 4.0, 6 * (5*alpha - 11) / ((alpha - 3) * (alpha - 4)), jnp.nan)
        return self._scalar_transform({
            "mean": mean, 
            "mode": mode, 
            "variance": variance, 
            "std": std, 
            "skewness": skewness, 
            "kurtosis": kurtosis})
    
    # fitting
    def _fit_mle(self, x: ArrayLike, lr: float, maxiter: int) -> dict:
        key1, key2 = random.split(DEFAULT_RANDOM_KEY)

        gamma_params: dict = gamma.example_params()
        params0: jnp.ndarray = jnp.array([gamma.rvs(size=(), key=key1, params=gamma_params), 
                                        gamma.rvs(size=(), key=key2, params=gamma_params)])
        
        res = projected_gradient(
            f=self._mle_objective, x0=params0, x=x, lr=lr, maxiter=maxiter, 
            projection_method='projection_non_negative')
        
        alpha, beta = res["x"]
        return self._params_dict(alpha=alpha, beta=beta)#, res["fun"]
    
    def fit(self, x: ArrayLike, lr: float = 0.1, maxiter: int = 100) -> dict:
        x: jnp.ndarray = _univariate_input(x)[0]
        return self._fit_mle(x=x, lr=lr, maxiter=maxiter)
    

ig = IG("IG")


"""File containing the copuLAX implementation of the skewed-T distribution."""
import jax.numpy as jnp
from jax import lax, custom_vjp, random
from jax._src.typing import ArrayLike, Array
from copy import deepcopy

from copulax._src._distributions import Univariate
from copulax._src.typing import Scalar
from copulax._src.univariate._utils import _univariate_input
from copulax.special import kv
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.univariate._cdf import _cdf, cdf_bwd, _cdf_fwd
from copulax._src.optimize import projected_gradient
from copulax._src.univariate.student_t import student_t
from copulax._src.univariate.ig import ig
from copulax._src.univariate._mean_variance import mean_variance_ldmle_params, mean_variance_stats
from copulax._src.univariate._rvs import mean_variance_sampling


class SkewedTBase(Univariate):
    @classmethod
    def _params_dict(cls, nu: Scalar, mu: Scalar, sigma: Scalar, 
                     gamma: Scalar) -> dict:
        d: dict = {"nu": nu, "mu": mu, "sigma": sigma, "gamma": gamma}
        return cls._args_transform(d)
    
    @staticmethod
    def _params_to_tuple(params: dict) -> tuple:
        params = SkewedTBase._args_transform(params)
        return params["nu"], params["mu"], params["sigma"], params["gamma"]
    
    @staticmethod
    def _params_to_array(params: dict) -> Array:
        return jnp.asarray(SkewedTBase._params_to_tuple(params)).flatten()

    @classmethod
    def _support(cls, *args, **kwargs) -> tuple:
        return -jnp.inf, jnp.inf  

    def example_params(self, *args, **kwargs) -> dict:
        r"""Example parameters for the skewed-T distribution.
        
        This is a four parameter family, with the skewed-T being defined 
        by its degrees of freedom `nu`, location `mu`, scale `sigma` and 
        skewness `gamma`. It is a generalisation of the student-T 
        distribution, which it includes as a special case when gamma is 
        zero. Here, we adopt the parameterization used by McNeil et al. 
        (2005).
        """
        return self._params_dict(nu=4.5, mu=0.0, sigma=1.0, gamma=1.0)
    
    @staticmethod
    def _skewed_stable_logpdf(stability: float, x: ArrayLike, params: dict) -> Array:
        nu, mu, sigma, gamma = SkewedTBase._params_to_tuple(params)
        x, xshape = _univariate_input(x)

        s: float = 0.5 * (nu + 1)
        c: float = (jnp.log(2.0) * (1 - s) 
                    - lax.lgamma(0.5 * nu) 
                    - 0.5 * jnp.log(jnp.pi * nu + stability) 
                    - jnp.log(sigma + stability))

        P: jnp.ndarray = (x - mu) * lax.pow(sigma, -2)
        Q: jnp.ndarray = P * (x - mu)
        R: jnp.ndarray = lax.pow(gamma / sigma, 2)

        T: jnp.ndarray = (jnp.log(kv(s, lax.sqrt((nu + Q) * R)) + stability) 
                          + P * gamma)
        B: jnp.ndarray = (-s * 0.5 * jnp.log((nu + Q) * R + stability) 
                          + s * jnp.log(1 + Q / (nu + stability)))

        logpdf: jnp.ndarray = c + T - B
        return logpdf.reshape(xshape)

    @classmethod
    def _stable_logpdf(cls, x: ArrayLike, params: dict, stability: float = 0.0) -> Array:
        gamma: Scalar = params["gamma"]
        return lax.cond(gamma == 0, 
                        lambda x: student_t._stable_logpdf(stability=stability,x=x, params=params), 
                        lambda x: cls._skewed_stable_logpdf(x=x, params=params, stability=stability), x)
    # @staticmethod
    # def _unnormalized_pdf(x: ArrayLike, params: dict, stability: float = 0.0) -> Array:
    #     return jnp.exp(SkewedTBase._unnormalized_logpdf(x=x, params=params, stability=stability))
    
    @staticmethod
    def logpdf(x: ArrayLike, params: dict) -> Array:
        params = SkewedTBase._args_transform(params)
        # normalising_constant: float = _cdf(dist=SkewedTBase, x=jnp.inf, params=params)
        # return SkewedTBase._unnormalized_logpdf(x=x, params=params)# - jnp.log(normalising_constant)
        return SkewedTBase._stable_logpdf(x=x, params=params, stability=1e-30)
    
    @staticmethod
    def pdf(x: ArrayLike, params: dict) -> Array:
        return jnp.exp(SkewedTBase.logpdf(x=x, params=params))
    
    # ppf
    # def _get_x0(self, params: dict) -> Scalar:
    #     return self._args_transform(params)["mu"]
    
    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, key: Array = DEFAULT_RANDOM_KEY) -> Array:
        nu, mu, sigma, gamma = self._params_to_tuple(params)
        
        key1, key2 = random.split(key)
        W: jnp.ndarray = ig.rvs(size=size, key=key1, params={'alpha': nu*0.5, 'beta': nu*0.5})
        return mean_variance_sampling(key=key2, W=W, shape=size, mu=mu, sigma=sigma, gamma=gamma)
    
    # stats
    def _get_w_stats(self, nu: float) -> dict:
        ig_params: dict = {'alpha': nu * 0.5, 'beta': nu * 0.5}
        ig_stats: dict = ig.stats(params=ig_params)
        w_mean =  jnp.where(jnp.isnan(ig_stats['mean']), ig_stats['mode'], ig_stats['mean'])
        w_variance = jnp.where(jnp.isnan(ig_stats['variance']), jnp.var(ig.rvs(size=100, params=ig_params)), ig_stats['variance'])
        return {'mean': w_mean, 'variance': w_variance}

    def stats(self, params: dict) -> dict:
        nu, mu, sigma, gamma = self._params_to_tuple(params)
        w_stats: dict = self._get_w_stats(nu)
        return self._scalar_transform(mean_variance_stats(mu=mu, sigma=sigma, gamma=gamma, w_stats=w_stats))
    
    # fitting
    # def _mle_objective(self, params_arr: jnp.ndarray, x: jnp.ndarray, 
    #                    *args, **kwargs) -> Scalar:
    #     # overriding base method to use unnormalized-loglikelihood for faster iterations
    #     params: dict = self._params_from_array(params_arr, *args, **kwargs)
    #     return self._unnormalized_logpdf(x=x, params=params, stability=1e-30).sum()

    def _fit_mle(self, x: jnp.ndarray, lr: float, maxiter: int) -> dict:
        eps: float = 1e-8
        constraints: tuple = (jnp.array([[eps, -jnp.inf, eps, -jnp.inf]]).T, 
                            jnp.array([[jnp.inf, jnp.inf, jnp.inf, jnp.inf]]).T)
        
        projection_options: dict = {'hyperparams': constraints}

        key1, key2 = random.split(DEFAULT_RANDOM_KEY)
        params0: jnp.ndarray = jnp.array([
            jnp.abs(random.normal(key1, ())), 
            x.mean(),
            x.std(),
            random.normal(key2, ())])
        
        params0: jnp.ndarray = jnp.array([
            1.0, 
            x.mean(),
            x.std(),
            1.0])
        
        res: dict = projected_gradient(f=self._mle_objective, x0=params0,
                                    projection_method='projection_box', 
                                    projection_options=projection_options, 
                                    x=x, lr=lr, maxiter=maxiter)
        nu, mu, sigma, gamma = res['x']
        return self._params_dict(nu=nu, mu=mu, sigma=sigma, gamma=gamma)#, res['fun']

    def _ldmle_objective(self, params: jnp.ndarray, x: jnp.ndarray, sample_mean: Scalar, sample_variance: Scalar) -> jnp.ndarray:
        nu, gamma = params
        ig_stats: dict = self._get_w_stats(nu=nu)
        mu, sigma = mean_variance_ldmle_params(
            stats=ig_stats, gamma=gamma, sample_mean=sample_mean, 
            sample_variance=sample_variance)
        return self._mle_objective(params_arr=jnp.array([nu, mu, sigma, gamma]), x=x)
    
    def _fit_ldmle(self, x: jnp.ndarray, lr: float, maxiter: int) -> dict:
        eps: float = 1e-8
        min_nu: float = 4.0 
        constraints: tuple = (jnp.array([[min_nu + eps, -jnp.inf]]).T, 
                            jnp.array([[jnp.inf, jnp.inf]]).T)

        key1, key2 = random.split(DEFAULT_RANDOM_KEY)
        params0: jnp.ndarray = jnp.array([min_nu+jnp.abs(random.normal(key1, ())), 
                                        random.normal(key2, ())])
        
        projection_options: dict = {'hyperparams': constraints}

        sample_mean, sample_variance = x.mean(), x.var()
        res: dict = projected_gradient(
            f=self._ldmle_objective, x0=params0, projection_method='projection_box', 
            projection_options=projection_options, x=x, sample_mean=sample_mean, 
            sample_variance=sample_variance, lr=lr, maxiter=maxiter)
        nu, gamma = res['x']
        ig_stats: dict = self._get_w_stats(nu=nu)
        mu, sigma = mean_variance_ldmle_params(
            stats=ig_stats, gamma=gamma, sample_mean=sample_mean, 
            sample_variance=sample_variance)
        return self._params_dict(nu=nu, mu=mu, sigma=sigma, gamma=gamma)#, res['fun']

    def fit(self, x: ArrayLike, method: str = 'LDMLE', 
            lr=0.1, maxiter: int = 100) -> dict:
        r"""Fit the distribution to the input data.

        Note:
            If you intend to jit wrap this function, ensure that 'method' is a 
            static argument.
        
        Args:
            x (ArrayLike): The input data to fit the distribution to.
            method (str): The fitting method to use.  Options are 
            'MLE' for maximum likelihood estimation, and 'LDMLE' for low-dimensional 
            maximum likelihood estimation. Defaults to 'LDMLE'. 
            kwargs: Additional keyword arguments to pass to the fit method.
        
        Returns:
            dict: The fitted distribution parameters.
        """ 
        x = _univariate_input(x)[0]
        if method == 'MLE':
            return self._fit_mle(x=x, lr=lr, maxiter=maxiter)
        else:
            return self._fit_ldmle(x=x, lr=lr, maxiter=maxiter)
        
    # cdf
    @staticmethod
    def _params_from_array(params_arr: jnp.ndarray, *args, **kwargs) -> dict:
        nu, mu, sigma, gamma = params_arr
        return SkewedTBase._params_dict(nu=nu, mu=mu, sigma=sigma, gamma=gamma)

    # @staticmethod
    # def _pdf_for_cdf(x: ArrayLike, *params_tuple) -> Array:
    #     params_array: jnp.ndarray = jnp.asarray(params_tuple).flatten()
    #     params: dict = SkewedTBase._params_from_array(params_array)
    #     return SkewedTBase._unnormalized_pdf(x=x, params=params)


def _vjp_cdf(x: ArrayLike, params: dict) -> Array:
    params = SkewedTBase._args_transform(params)
    return _cdf(dist=SkewedTBase, x=x, params=params)


_vjp_cdf_copy = deepcopy(_vjp_cdf)
_vjp_cdf = custom_vjp(_vjp_cdf)


def cdf_fwd(x: ArrayLike, params: dict) -> tuple[Array, tuple]:
    params = SkewedTBase._args_transform(params)
    return _cdf_fwd(dist=SkewedTBase, cdf_func=_vjp_cdf_copy, x=x, params=params)


_vjp_cdf.defvjp(cdf_fwd, cdf_bwd)


class SkewedT(SkewedTBase):
    r"""The skewed-t distribution is a generalisation of the continuous Student's 
    t-distribution that allows for skewness. It can also be expressed as a limiting 
    case of the Generalized Hyperbolic distribution when phi -> 0 in addition to 
    lambda = -0.5*chi.

    We use the 4 parameter McNeil et al (2005) specification of the distribution.
    """
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        return _vjp_cdf(x=x, params=params)
    

skewed_t = SkewedT("Skewed-T")


"""File containing the copulAX implementation of the continuous uniform 
distribution."""
import jax.numpy as jnp
from jax import lax, random
from jax._src.typing import ArrayLike, Array

from copulax._src._distributions import Univariate
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.typing import Scalar


class Uniform(Univariate):
    r"""The continuous uniform distribution.
    
    The continuous uniform distribution is defined as:
    
    .. math::
    
        f(x|a, b) = \frac{1}{b - a}
    
    where :math:`a` is the lower bound of the distribution and :math:`b` is the 
    upper bound.
    """
    @classmethod
    def _params_dict(cls, a: Scalar, b: Scalar) -> dict:
        d: dict = {"a": a, "b": b}
        return cls._args_transform(d)
    
    @classmethod
    def _params_to_tuple(cls, params: dict) -> tuple:
        params = cls._args_transform(params)
        return params["a"], params["b"]
    
    def example_params(self, *args, **kwargs) -> dict:
        r"""Example parameters for the uniform distribution.
        
        This is a two parameter family, with the uniform being defined by 
        its lower and upper bounds.
        """
        return self._params_dict(a=0.0, b=1.0)
    
    @classmethod
    def _support(cls, params: dict) -> tuple:
        a, b = cls._params_to_tuple(params)
        return a, b

    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        a, b = self._params_to_tuple(params)

        log_pdf: jnp.ndarray = -jnp.log(lax.sub(b, a))
        log_pdf = jnp.where(jnp.logical_and(x >= a, x <= b), log_pdf, -jnp.inf)
        return log_pdf.reshape(xshape)
    
    def logcdf(self, x: ArrayLike, params : dict) -> Array:
        x, xshape = _univariate_input(x)
        a, b = self._params_to_tuple(params)

        log_cdf: jnp.ndarray = jnp.log(lax.sub(x, a)) - jnp.log(lax.sub(b, a))
        log_cdf = jnp.where(jnp.logical_and(x >= a, x <= b), log_cdf, -jnp.inf)
        return log_cdf.reshape(xshape)
    
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        return jnp.exp(self.logcdf(x=x, params=params))
    
    # ppf
    # def _get_x0(self, params: dict) -> Scalar:
    #     stats: dict = self.stats(params=params)
    #     return stats["mean"]
    
    def _ppf(self, q: ArrayLike, params: dict, *args, **kwargs) -> Array:
        q, qshape = _univariate_input(q)
        a, b = self._params_to_tuple(params)
        
        ppf_values: jnp.ndarray = lax.add(a, lax.mul(q, lax.sub(b, a)))
        ppf_values = jnp.where(jnp.logical_and(q >= 0, q <= 1), ppf_values, jnp.nan)
        return ppf_values.reshape(qshape)
    
    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, key=DEFAULT_RANDOM_KEY) -> Array:
        a, b = self._params_to_tuple(params)
        return random.uniform(key=key, shape=size, minval=a, maxval=b)
    
    # stats
    def stats(self, params: dict) -> dict:
        a, b = self._params_to_tuple(params)

        mean: Scalar = (a + b) / 2
        variance: Scalar = lax.pow(b - a, 2) / 12
        std: Scalar = jnp.sqrt(variance)
        return self._scalar_transform({
            'mean': mean, 
            'median': mean, 
            'variance': variance,
            'std': std, 
            'skewness': 0.0, 
            'kurtosis': -6 / 5})

    # fitting
    def fit(self, x: ArrayLike, *args, **kwargs) -> dict:
        x: jnp.ndarray = _univariate_input(x)[0]
        a: Scalar = jnp.min(x)
        b: Scalar = jnp.max(x)
        return self._params_dict(a=a, b=b)
    

uniform = Uniform("Uniform")


"""Contains utility functions for the copulax package."""
from jax import random
import jax.numpy as jnp
import requests
import os
import sys
import logging


def _type_check_pos_int(value: int, name: str) -> None:
    """Check if the value is an integer."""
    if not isinstance(value, int):
        raise TypeError(f"{name} must be an integer.")
    if value <= 0:
        raise ValueError(f"{name} must be a positive integer.")


###############################################################################
# Random Key Generation
###############################################################################
def get_api_random_key(max_attempts: int = 10) -> random.key:
    """Returns a random key for use in JAX functions.

    This function connects to random.org to obtain a true random seed 
    using atmospheric noise. This involves making a GET request to the 
    random.org API. This can be slow and if being called many times in a 
    loop, may not be desirable. It also involves using a third party 
    source for randomness, which may not be desirable for all 
    applications.

    Note:
        https://www.random.org/
        Not jitable.
        No gradients.

    Args:
        max_attempts (int, optional): The maximum number of attempts to 
            connect to random.org to obtain a true random seed. If a 
            connection cannot be established after this threshold, an 
            error is raised.

    Returns:
        A random key for use in JAX functions.
    """
    _type_check_pos_int(max_attempts, "max_attempts")
    # attempt to connect to random.org to obtain a true random seed
    url: str = "https://www.random.org/integers/?num=1&min=-1000000000&max=1000000000&col=1&base=10&format=plain&rnd=new"
    attempts: int = 0
    seed = None
    while attempts < max_attempts:
        response = requests.get(url)
        if response.status_code == 200:
            seed: int = int(response.text)
            break
        attempts += 1
    
    if seed is None:
        raise ConnectionError(f"Unable to connect to random.org after max = {max_attempts} attempts.")
    return random.key(seed)


def get_local_random_key(bytestring_size: int = 7) -> random.key:
    """Returns a random key for use in JAX functions.

    This function uses a hardware based approach to generate a random 
    seed. This involves generating random bytes using os.urandom and 
    converting the byte string to an integer. This method is usually 
    faster than the API method and can be used if reliance on a third 
    party is not desired. Quality of randomness is dependent upon the OS 
    implementation.

    Args:
        bytestring_size (int, optional): The length of the byte string 
            to generate using os.urandom. If the resultant integer is 
            outside the bounds of int64, the integer is truncated to fit 
            within the bounds as to be compatible with jax. 
            The default is 7.

    Note:
        Not jitable.
        No gradients.
    
    Returns:
        A random key for use in JAX functions.
    """
    _type_check_pos_int(bytestring_size, "bytestring_size")
    # generating raw seed int from random bytes
    byte_str: str = os.urandom(bytestring_size)
    seed: int = int.from_bytes(bytes=byte_str, byteorder=sys.byteorder, 
                               signed=True)

    # ensuring the seed is within the bounds of int64
    int64_bounds = jnp.iinfo(jnp.int64)
    if not (int64_bounds.min <= seed <= int64_bounds.max):
        # generated integer is outside the bounds of int64, hence truncating
        seed_str: str = str(seed)
        relevant_bound = int64_bounds.max if seed > 0 else int64_bounds.min
        truncated_length = len(str(relevant_bound)) - 1
        seed = int(seed_str[:truncated_length])
    return random.key(seed)


def get_random_key(local: bool = True, bytestring_size: int = 7, 
                   max_attempts: int = 10) -> random.key:
    """Returns a random key for use in JAX functions.

    Wrapper for the different copulAX for generating a random key. These are:
        1. get_local_random_key; Use a hardware based approach to 
        generate a random seed. This is the default method and involves 
        generating random bytes using os.urandom and converting the byte 
        string to an integer. This method is usually faster than the API 
        method and can be used if reliance on a third party is not 
        desired. Quality of randomness is dependent upon the OS 
        implementation.
        2. get_api_random_key; An API based approach which connects to 
        random.org to obtain a true random seed using atmospheric noise. 
        This involves making a GET request to the random.org API. This 
        can be slow and if being called many times in a loop, may not be 
        desirable. It also involves using a third party source for 
        randomness, which may not be desirable for all applications.

    See Also:
        get_local_random_key, get_api_random_key

    Note:
        https://www.random.org/
        Not jitable.
        No gradients.

    Args:
        local (bool, optional): If True, use the local method to 
            generate a random key. If False, use the API method to 
            generate a random key.
        bytestring_size (int, optional): Only used if local is True. The 
            length of the byte string to generate using os.urandom. If 
            the resultant integer is outside the bounds of int64, the 
            integer is truncated to fit within the bounds as to be 
            compatible with jax. The default is 7.
        max_attempts (int, optional): Only used if local is False. The 
            maximum number of attempts to connect to random.org to 
            obtain a true random seed. If a connection cannot be 
            established after this threshold, the local generation 
            method is used. The default is 10.
    
    Returns:
        A random key for use in JAX functions.
    """
    if local:
        # use the local method to generate a random key
        return get_local_random_key(bytestring_size=bytestring_size)
    
    # use the API method to generate a random key
    try:
        return get_api_random_key(max_attempts=max_attempts)
    except ConnectionError as e:
        msg: str = f"{str(e)}\nUsing local method instead."
        logging.warning(msg)
        return get_local_random_key(bytestring_size=bytestring_size)


DEFAULT_RANDOM_KEY = get_local_random_key()


from copulax._src.multivariate._shape import corr, cov, random_correlation, random_covariance
from copulax._src.multivariate.mvt_normal import mvt_normal
from copulax._src.multivariate.mvt_student_t import mvt_student_t
from copulax._src.multivariate.mvt_gh import mvt_gh
from copulax._src.multivariate.mvt_skewed_t import mvt_skewed_t


r"""Contains collections of univariate distributions.

Currently the following distribution collections are implemented:
- distributions: contains all implemented distributions, grouped by data type.
- common: contains the most common distributions, grouped by data type.
- continuous: contains all continuous distributions.
- discrete: contains all discrete distributions.
"""
# importing pytree distribution objects
from copulax._src.univariate.gamma import gamma
from copulax._src.univariate.gh import gh
from copulax._src.univariate.gig import gig
from copulax._src.univariate.ig import ig
from copulax._src.univariate.lognormal import lognormal
from copulax._src.univariate.normal import normal
from copulax._src.univariate.skewed_t import skewed_t
from copulax._src.univariate.student_t import student_t
from copulax._src.univariate.uniform import uniform

# importing categorizing map
from copulax._src._distributions import dist_map

__dists = (gamma, gh, gig, ig, lognormal, normal, skewed_t, student_t, uniform)
__common_names = ('Gamma', 'Normal', 'Student-T', 'Uniform', 'LogNormal',)

_all_dists: list = []
_dist_tree = {'continuous': {}, 'discrete': {}}
_dist_tree['common'] = _dist_tree.copy()
_all_dist_objects: list = []
for dist in __dists:
    _all_dists.append(dist.name)
    _all_dist_objects.append(dist_map.name_map[dist.name]['object_name'])

    _dist_tree[dist.name] = dist

    if dist.name in __common_names:
        _dist_tree['common'][dist.dtype][dist.name] = dist
    
    _dist_tree[dist.dtype][dist.name] = dist

_all_dists: tuple = tuple(_all_dists)
distributions = _dist_tree.copy()
common: dict = distributions.pop('common')
continuous: dict = distributions['continuous'].copy()
discrete: dict = distributions['discrete'].copy()

__all__ = ['distributions', 'common', 'continuous', 'discrete']
__all__ += _all_dist_objects
